import numpy as np
from visdom import Visdom, _assert_opts, _opts2layout


class ExtraVisdom(Visdom):
    def histogram_3d(self, X, win=None, env=None, opts=None):
        """
        Given an array it plots the histrograms of the entries.

        Args:
            X : An array of at least 2 dimensions, where the first dimensions gives the number of histograms.
            win: Window name.
            env: Env name.
            opts: dict with options, especially opts['numbins'] (number of histogram bins) and opts['mutiplier']
        ( factor to stretch / queeze the values on the x axis) should be considered.

        Returns:
            The send result.
        """

        opts = {} if opts is None else opts
        _assert_opts(opts)

        X = np.asarray(X)
        assert X.ndim >= 2, 'X must have atleast 2 dimensions'

        opts['numbins'] = opts.get('numbins', min(30, len(X[0])))
        opts['mutiplier'] = opts.get('numbins', 100)

        traces = []
        for i, array in enumerate(X):
            array = array.flatten()
            bins, intervals = np.histogram(array, bins=opts['numbins'])

            x = []
            y = []
            z = []
            prev_interv = 0.
            for b, iv in zip(bins, intervals):
                interval_middle = float((prev_interv + iv) / 2.) * opts['mutiplier']
                z.append([float(b), float(b)])
                y.append([interval_middle, interval_middle])
                x.append([i * 2, i * 2 + 0.5])
                prev_interv = float(iv)
            traces.append(dict(
                z=z,
                x=x,
                y=y,
                # colorscale=[[i, 'rgb(%d,%d,255)' % (ci, ci)] for i in np.arange(0, 1.1, 0.1)],
                # autocolorscale=True,
                showscale=False,
                type='surface',
            ))

        return self._send({
            'data': traces,
            'win': win,
            'eid': env,
            'layout': _opts2layout(opts),
            'opts': opts,
        })



from __future__ import print_function

import atexit
import fnmatch
import os
import warnings
from multiprocessing import Process

import torch
import numpy as np

from trixi.logger.abstractlogger import threaded
from trixi.logger.experiment import ExperimentLogger
from trixi.logger.file.pytorchplotfilelogger import PytorchPlotFileLogger
from trixi.util import name_and_iter_to_filename
from trixi.util.pytorchutils import update_model, get_vanilla_image_gradient, get_guided_image_gradient, \
    get_smooth_image_gradient


class PytorchExperimentLogger(ExperimentLogger):
    """
    A single class for logging your pytorch experiments to file.
    Extends the ExperimentLogger also also creates a experiment folder with a file structure:

    The folder structure is :
        base_dir/
            new_experiment_folder/
                checkpoint/
                config/
                img/
                log/
                plot/
                result/
                save/


    """

    def __init__(self, *args, **kwargs):
        """Initializes the PytorchExperimentLogger and parses the arguments to the ExperimentLogger"""

        super(PytorchExperimentLogger, self).__init__(*args, **kwargs)
        self.plot_logger = PytorchPlotFileLogger(self.img_dir, self.plot_dir)

    def show_images(self, images, name, **kwargs):
        """
        Saves images in the img folder

        Args:
            images: The images to be saved
            name: file name of the new image file

        """
        self.plot_logger.show_images(images, name, **kwargs)

    def show_image_grid(self, image, name, **kwargs):
        """
        Saves images in the img folder as a image grid

        Args:
            images: The images to be saved
            name: file name of the new image file

        """
        self.plot_logger.show_image_grid(image, name, **kwargs)

    def show_image_grid_heatmap(self, heatmap, background=None, name="heatmap", **kwargs):
        """
        Saves images in the img folder as a image grid

        Args:
            heatmap: The images to be converted to a heatmap
            background: Context of the heatmap (to be underlayed)
            name: file name of the new image file

        """
        self.plot_logger.show_image_grid_heatmap(heatmap=heatmap, background=background, name=name, **kwargs)

    @staticmethod
    @threaded
    def save_model_static(model, model_dir, name):
        """
        Saves a pytorch model in a given directory (using pytorch)

        Args:
            model: The model to be stored
            model_dir: The directory in which the model file should be written
            name: The file name of the model file

        """

        model_file = os.path.join(model_dir, name)
        torch.save(model.state_dict(), model_file)

    def save_model(self, model, name, n_iter=None, iter_format="{:05d}", prefix=False):
        """
        Saves a pytorch model in the model directory of the experiment folder

        Args:
            model: The model to be stored
            name: The file name of the model file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix

        """

        if n_iter is not None:
            name = name_and_iter_to_filename(name,
                                             n_iter,
                                             ".pth",
                                             iter_format=iter_format,
                                             prefix=prefix)

        if not name.endswith(".pth"):
            name += ".pth"

        self.save_model_static(model=model,
                               model_dir=self.checkpoint_dir,
                               name=name)

    @staticmethod
    @threaded
    def load_model_static(model, model_file, exclude_layers=(), warnings=True):
        """
        Loads a pytorch model from a given directory (using pytorch)


        Args:
            model: The model to be loaded (whose parameters should be restored)
            model_file: The file from which the model parameters should be loaded
            exclude_layers: List of layer names which should be excluded from restoring
            warnings (bool): Flag which indicates if method should warn if not everything went perfectly

        """

        if os.path.exists(model_file):

            pretrained_dict = torch.load(model_file, map_location=lambda storage, loc: storage)
            update_model(model, pretrained_dict, exclude_layers, warnings)

        else:

            raise IOError("Model file does not exist!")

    def load_model(self, model, name, exclude_layers=(), warnings=True):
        """
        Loads a pytorch model from the model directory of the experiment folder


        Args:
            model: The model to be loaded (whose parameters should be restored)
            name: The file name of the model file
            exclude_layers: List of layer names which should be excluded from restoring
            warnings: Flag which indicates if method should warn if not everything went perfectlys


        """

        if not name.endswith(".pth"):
            name += ".pth"

        self.load_model_static(model=model,
                               model_file=os.path.join(self.checkpoint_dir, name),
                               exclude_layers=exclude_layers,
                               warnings=warnings)

    @staticmethod
    @threaded
    def save_checkpoint_static(checkpoint_dir, name, move_to_cpu=False, **kwargs):
        """
        Saves a checkpoint/dict in a given directory (using pytorch)

        Args:
            checkpoint_dir: The directory in which the checkpoint file should be written
            name: The file name of the checkpoint file
            move_to_cpu (bool): Flag, if all pytorch tensors should be moved to cpu before storing
            **kwargs: dict which is actually saved

        """
        for key, value in kwargs.items():
            if isinstance(value, torch.nn.Module) or isinstance(value, torch.optim.Optimizer):
                kwargs[key] = value.state_dict()

        checkpoint_file = os.path.join(checkpoint_dir, name)

        def to_cpu(obj):
            if hasattr(obj, "cpu"):
                return obj.cpu()
            elif isinstance(obj, dict):
                return {key: to_cpu(val) for key, val in obj.items()}
            else:
                return obj

        if move_to_cpu:
            torch.save(to_cpu(kwargs), checkpoint_file)
        else:
            torch.save(kwargs, checkpoint_file)

    def save_checkpoint(self, name, n_iter=None, iter_format="{:05d}", prefix=False, **kwargs):
        """
        Saves a checkpoint in the checkpoint directory of the experiment folder

        Args:
            name: The file name of the checkpoint file
            n_iter: The iteration number, formatted with the iter_format and added to the checkpoint name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            **kwargs:  dict which is actually saved (key=name, value=variable to be stored)

        """

        if n_iter is not None:
            name = name_and_iter_to_filename(name,
                                             n_iter,
                                             ".pth.tar",
                                             iter_format=iter_format,
                                             prefix=prefix)

        if not name.endswith(".pth.tar"):
            name += ".pth.tar"

        self.save_checkpoint_static(self.checkpoint_dir, name=name, **kwargs)

    @staticmethod
    def load_checkpoint_static(checkpoint_file, exclude_layer_dict=None, warnings=True, **kwargs):
        """
        Loads a checkpoint/dict in a given directory (using pytorch)

        Args:
            checkpoint_file: The checkpoint from which the checkpoint/dict should be loaded
            exclude_layer_dict: A dict with key 'model_name' and a list of all layers of 'model_name' which should
            not be restored
            warnings: Flag which indicates if method should warn if not everything went perfectlys
            **kwargs: dict which is actually loaded (key=name (used to save the checkpoint) , value=variable to be
            loaded/ overwritten)

        Returns: The kwargs dict with the loaded/ overwritten values

        """

        if exclude_layer_dict is None:
            exclude_layer_dict = {}

        checkpoint = torch.load(checkpoint_file, map_location=lambda storage, loc: storage)

        for key, value in kwargs.items():
            if key in checkpoint:
                if isinstance(value, torch.nn.Module) or isinstance(value, torch.optim.Optimizer):
                    exclude_layers = exclude_layer_dict.get(key, [])
                    update_model(value, checkpoint[key], exclude_layers, warnings)
                else:
                    kwargs[key] = checkpoint[key]

        return kwargs

    def load_checkpoint(self, name, exclude_layer_dict=None, warnings=True, **kwargs):
        """
        Loads a checkpoint from the checkpoint directory of the experiment folder

        Args:
            name: The name of the checkpoint file
            exclude_layer_dict: A dict with key 'model_name' and a list of all layers of 'model_name' which should
            not be restored
            warnings: Flag which indicates if method should warn if not everything went perfectlys
            **kwargs: dict which is actually loaded (key=name (used to save the checkpoint) , value=variable to be
            loaded/ overwritten)

        Returns: The kwargs dict with the loaded/ overwritten values

        """

        if not name.endswith(".pth.tar"):
            name += ".pth.tar"

        checkpoint_file = os.path.join(self.checkpoint_dir, name)
        return self.load_checkpoint_static(checkpoint_file=checkpoint_file,
                                           exclude_layer_dict=exclude_layer_dict,
                                           warnings=warnings,
                                           **kwargs)

    def save_at_exit(self, name="checkpoint_end", **kwargs):
        """
        Saves a dict as checkpoint if the program exits (not garanteed to work 100%)

        Args:
            name: Name of the checkpoint file
            **kwargs: dict which is actually saved (key=name, value=variable to be stored)

        """

        if not name.endswith(".pth.tar"):
            name += ".pth.tar"

        def save_fnc():
            self.save_checkpoint(name, **kwargs)
            print("Checkpoint saved securely... =)")

        atexit.register(save_fnc)

    def get_save_checkpoint_fn(self, name="checkpoint", **kwargs):
        """
        A function which returns a function which takes n_iter as arguments and saves the current values of the
        variables given as kwargs as a checkpoint file.


        Args:
            name: Base-name of the checkpoint file
            **kwargs:  dict which is actually saved, when the returned function is called

        Returns: Function which takes n_iter as arguments and saves a checkpoint file
        """

        def save_fnc(n_iter, iter_format="{:05d}", prefix=False):
            self.save_checkpoint(name=name,
                                 n_iter=n_iter,
                                 iter_format=iter_format,
                                 prefix=prefix,
                                 **kwargs)

        return save_fnc

    @staticmethod
    def load_last_checkpoint_static(dir_, name=None, **kwargs):
        """
        Loads the (alphabetically) last checkpoint file in a given directory

        Args:
            dir_: The directory to look for the (alphabetically) last checkpoint
            name: String pattern which indicates the files to look form
            **kwargs: dict which is actually loaded (key=name (used to save the checkpoint) , value=variable to be
            loaded/ overwritten)

        Returns:  The kwargs dict with the loaded/ overwritten values

        """

        if name is None:
            name = "*checkpoint*.pth.tar"

        checkpoint_files = []

        for root, dirs, files in os.walk(dir_):
            for filename in fnmatch.filter(files, name):
                checkpoint_file = os.path.join(root, filename)
                checkpoint_files.append(checkpoint_file)

        if len(checkpoint_files) == 0:
            return None

        last_file = sorted(checkpoint_files, reverse=True)[0]

        return PytorchExperimentLogger.load_checkpoint_static(last_file, **kwargs)

    def load_last_checkpoint(self, **kwargs):
        """
                Loads the (alphabetically) last checkpoint file in the checkpoint directory in the experiment folder

                Args:
                    **kwargs: dict which is actually loaded (key=name (used to save the checkpoint) , value=variable to be
                    loaded/ overwritten)

                Returns:  The kwargs dict with the loaded/ overwritten values

                """
        return self.load_last_checkpoint_static(self.checkpoint_dir, **kwargs)

    def print(self, *args):
        """
        Prints the given arguments using the text logger print function

        Args:
            *args: Things to be printed

        """
        self.text_logger.print(*args)

    @staticmethod
    def get_roc_curve(tensor, labels, reduce_to_n_samples=None, use_sub_process=False, results_fn=lambda
            x, *y, **z: None):
        """
        Displays a roc curve given a tensor with scores and the coresponding labels

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            reduce_to_n_samples: Reduce/ downsample to to n samples for fewer data points
            use_sub_process: Use a sub process to do the processing, if true nothing is returned
            results_fn: function which is called with the results/ return values. Expected f(tpr, fpr)

        """
        from sklearn import metrics

        def __get_roc_curve(tensor, labels, reduce_to_n_samples=None, results_fn=lambda x, *y, **z: None):

            if not isinstance(labels, list):
                labels = labels.flatten()
            if not isinstance(tensor, list):
                tensor = tensor.flatten()

            fpr, tpr, thresholds = metrics.roc_curve(labels, tensor)
            if reduce_to_n_samples is not None:
                fpr = [np.mean(x) for x in np.array_split(fpr, reduce_to_n_samples)]
                tpr = [np.mean(x) for x in np.array_split(tpr, reduce_to_n_samples)]
            results_fn(tpr, fpr)

            return tpr, fpr
            # self.show_lineplot(tpr, fpr, name=name, opts={"fillarea": True, "webgl": True})
            # self.add_to_graph(x_vals=np.arange(0, 1.1, 0.1), y_vals=np.arange(0, 1.1, 0.1), name=name, append=True)

        if use_sub_process:
            p = Process(target=__get_roc_curve, kwargs=dict(tensor=tensor,
                                                            labels=labels,
                                                            reduce_to_n_samples=reduce_to_n_samples,
                                                            results_fn=results_fn
                                                            ))
            atexit.register(p.terminate)
            p.start()
        else:
            try:
                return __get_roc_curve(tensor=tensor,
                                       labels=labels,
                                       reduce_to_n_samples=reduce_to_n_samples,
                                       results_fn=results_fn
                                       )
            except Exception as e:
                warnings.warn("Sth went wrong with calculating the roc curve")

    @staticmethod
    def get_pr_curve(tensor, labels, reduce_to_n_samples=None, use_sub_process=False,
                     results_fn=lambda x, *y, **z: None):
        """
        Displays a precision recall curve given a tensor with scores and the coresponding labels

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            reduce_to_n_samples: Reduce/ downsample to to n samples for fewer data points
            use_sub_process: Use a sub process to do the processing, if true nothing is returned
            results_fn: function which is called with the results/ return values. Expected f(precision, recall)

        """
        from sklearn import metrics

        def __get_pr_curve(tensor, labels, reduce_to_n_samples=None, results_fn=lambda x, *y, **z: None):

            if not isinstance(labels, list):
                labels = labels.flatten()
            if not isinstance(tensor, list):
                tensor = tensor.flatten()

            precision, recall, thresholds = metrics.precision_recall_curve(labels, tensor)
            if reduce_to_n_samples is not None:
                precision = [np.mean(x) for x in np.array_split(precision, reduce_to_n_samples)]
                recall = [np.mean(x) for x in np.array_split(recall, reduce_to_n_samples)]
            results_fn(precision, recall)

            return precision, recall
            # self.show_lineplot(precision, recall, name=name, opts={"fillarea": True, "webgl": True})
            # self.add_to_graph(x_vals=np.arange(0, 1.1, 0.1), y_vals=np.arange(0, 1.1, 0.1), name=name, append=True)

        if use_sub_process:
            p = Process(target=__get_pr_curve, kwargs=dict(tensor=tensor,
                                                           labels=labels,
                                                           reduce_to_n_samples=reduce_to_n_samples,
                                                           results_fn=results_fn
                                                           ))
            atexit.register(p.terminate)
            p.start()
        else:
            try:
                return __get_pr_curve(tensor=tensor,
                                      labels=labels,
                                      reduce_to_n_samples=reduce_to_n_samples,
                                      results_fn=results_fn
                                      )
            except Exception as e:
                warnings.warn("Sth went wrong with calculating the pr curve")

    @staticmethod
    def get_classification_metrics(tensor, labels, name="", metric=("roc-auc", "pr-score"), use_sub_process=False,
                                   tag_name=None, results_fn=lambda x, *y, **z: None):
        """
        Displays some classification metrics as line plots in a graph (similar to show value (also uses show value
        for the caluclated values))

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            name: The name of the window
            metric: List of metrics to calculate. Options are: roc-auc, pr-auc, pr-score, mcc, f1
            reduce_to_n_samples: Reduce/ downsample to to n samples for fewer data points
            tag_name: Name for the tag, if no given use name
            use_sub_process: Use a sub process to do the processing, if true nothing is returned
            results_fn: function which is called with the results/ return values. Expected f(val, name, tag)

        Returns:

        """

        from sklearn import metrics

        def __get_classification_metrics(tensor, labels, name="", metric=("roc-auc", "pr-score"),
                                         tag_name=None, results_fn=lambda x, *y, **z: None):

            vals = []
            tags = []

            if not isinstance(labels, list):
                labels = labels.flatten()
            if not isinstance(tensor, list):
                tensor = tensor.flatten()

            if "roc-auc" in metric:
                roc_auc = metrics.roc_auc_score(labels, tensor)
                vals.append(roc_auc)
                tags.append("roc-auc")
            if "pr-auc" in metric:
                precision, recall, thresholds = metrics.precision_recall_curve(labels, tensor)
                pr_auc = metrics.auc(recall, precision)
                vals.append(pr_auc)
                tags.append("pr-auc")
            if "pr-score" in metric:
                pr_score = metrics.average_precision_score(labels, tensor)
                vals.append(pr_score)
                tags.append("pr-score")
            if "mcc" in metric:
                mcc_score = metrics.matthews_corrcoef(labels, tensor)
                vals.append(mcc_score)
                tags.append("mcc")
            if "f1" in metric:
                f1_score = metrics.f1_score(labels, tensor)
                vals.append(f1_score)
                tags.append("f1")

            for val, tag in zip(vals, tags):
                results_fn(val, name=tag + "-" + name, tag=tag_name)

            return vals, tags

        if use_sub_process:
            p = Process(target=__get_classification_metrics, kwargs=dict(tensor=tensor,
                                                                         labels=labels,
                                                                         name=name,
                                                                         metric=metric,
                                                                         tag_name=tag_name,
                                                                         results_fn=results_fn
                                                                         ))
            atexit.register(p.terminate)
            p.start()
        else:
            try:
                return __get_classification_metrics(tensor=tensor,
                                                    labels=labels,
                                                    name=name,
                                                    metric=metric,
                                                    tag_name=tag_name,
                                                    results_fn=results_fn
                                                    )

            except Exception as e:
                warnings.warn("Sth went wrong with calculating the classification metrics")

    @staticmethod
    def get_input_gradient(model, inpt, err_fn, grad_type="vanilla", n_runs=20, eps=0.1,
                           abs=False, results_fn=lambda x, *y, **z: None):
        """
        Given a model creates calculates the error and backpropagates it to the image and saves it (saliency map).

        Args:
            model: The model to be evaluated
            inpt: Input to the model
            err_fn: The error function the evaluate the output of the model on
            grad_type: Gradient calculation method, currently supports (vanilla, vanilla-smooth, guided,
            guided-smooth) ( the guided backprob can lead to segfaults -.-)
            n_runs: Number of runs for the smooth variants
            eps: noise scaling to be applied on the input image (noise is drawn from N(0,1))
            abs (bool): Flag, if the gradient should be a absolute value
            results_fn: function which is called with the results/ return values. Expected f(grads)

        """
        model.zero_grad()

        if grad_type == "vanilla":
            grad = get_vanilla_image_gradient(model, inpt, err_fn, abs)
        elif grad_type == "guided":
            grad = get_guided_image_gradient(model, inpt, err_fn, abs)
        elif grad_type == "smooth-vanilla":
            grad = get_smooth_image_gradient(model, inpt, err_fn, n_runs, eps, grad_type="vanilla")
        elif grad_type == "smooth-guided":
            grad = get_smooth_image_gradient(model, inpt, err_fn, n_runs, eps, grad_type="guided")
        else:
            warnings.warn("This grad_type is not implemented yet")
            grad = torch.zeros_like(inpt)
        model.zero_grad()

        results_fn(grad)

        return grad

    def show_image_gradient(self, name, *args, **kwargs):
        """
        Given a model creates calculates the error and backpropagates it to the image and saves it.

        Args:
            name: Name of the file
            model: The model to be evaluated
            inpt: Input to the model
            err_fn: The error function the evaluate the output of the model on
            grad_type: Gradient calculation method, currently supports (vanilla, vanilla-smooth, guided,
            guided-smooth) ( the guided backprob can lead to segfaults -.-)
            n_runs: Number of runs for the smooth variants
            eps: noise scaling to be applied on the input image (noise is drawn from N(0,1))
            abs (bool): Flag, if the gradient should be a absolute value


        """
        grad = self.get_input_gradient(*args, **kwargs)
        self.show_image_grid(grad, name)



import json
import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from trixi.experiment import PytorchExperiment
from trixi.util import ResultLogDict


class TestPytorchExperiment(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.gettempdir()
        self.experiment = PytorchExperiment(name="test_experiment", base_dir=self.test_dir, n_epochs=10)

    def tearDown(self):
        self.experiment._exp_state = "Ended"
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_train(self):
        self.cntr = []
        self.experiment.train = lambda epoch: self.cntr.append(0)
        self.experiment.run()
        self.assertTrue(len(self.cntr) == 10, "Did not call train for the right number of epochs")

    def test_validate(self):
        self.cntr = []
        self.experiment.validate = lambda epoch: self.cntr.append(0)
        self.experiment.run()
        self.assertTrue(len(self.cntr) == 10, "Did not call train for the right number of epochs")

    def test_update_attributes(self):
        results2 = ResultLogDict("results-log.json", base_dir=self.test_dir)
        results2["test"] = 0
        self.experiment.test_var = "test"

        self.experiment.update_attributes({"results": results2, "test_var": "test2"})

        self.assertTrue(self.experiment.test_var == "test2", "Could not update simple var in update_attributes")
        self.assertTrue(self.experiment.results['test'] == 0, "Could not result dict var in update_attributes")

    def test_get_pytorch_modules(self):
        module = Net()
        self.experiment.net = module

        pyt_module = self.experiment.get_pytorch_modules()

        self.assertTrue("net" in pyt_module, "Name for module could not be correctly returned in get_pytorch_modules")
        self.assertTrue(pyt_module['net'] == module, "Module could not be correctly returned in  get_pytorch_modules")

    def test_get_pytorch_optimizers(self):
        optimizer = torch.optim.Adam(params=[torch.tensor(1)])
        self.experiment.optim = optimizer

        pyt_optim = self.experiment.get_pytorch_optimizers()

        self.assertTrue("optim" in pyt_optim, "Name for optimizer could not be correctly returned in "
                                              "get_pytorch_optimizers")
        self.assertTrue(pyt_optim['optim'] == optimizer, "Optimizer could not be correctly returned in  "
                                                         "get_pytorch_optimizers")

    def test_get_simple_vars(self):
        self.experiment.a = "test"
        self.experiment.b = 1
        self.experiment.c = True
        self.experiment.d = (1, 2, 3, 4)

        var = self.experiment.get_simple_variables()

        self.assertTrue("a" in var and "b" in var and "c" in var and "d" in var,
                        "Names for attributes could not be correctly returned in get_simple_variables")
        self.assertTrue(var['a'] == "test" and var['b'] == 1 and var['c'] == True and var['d'] == (1, 2, 3, 4),
                        "variables could not be correctly returned in get_simple_variables")

    def test_get_pytorch_vars(self):
        var = torch.tensor(1)
        self.experiment.pyt_var = var

        pyt_vars = self.experiment.get_pytorch_variables()

        self.assertTrue("pyt_var" in pyt_vars, "Name for variable could not be correctly returned in "
                                               "get_pytorch_variables")
        self.assertTrue(pyt_vars['pyt_var'] == var, "Variable could not be correctly returned in  "
                                                    "get_pytorch_variables")

    def test_save_results(self):
        self.experiment.add_result(0, "test")
        self.experiment.save_results("results-test.json")

        self.assertTrue(os.path.exists(os.path.join(self.experiment.elog.result_dir, "results-test.json")),
                        "result file could not be stored")

        with open(os.path.join(self.experiment.elog.result_dir, "results-test.json"), "r") as f:
            content = f.read()
            self.assertTrue("test" in content and "0" in content, "results content not sucessfully saved")

    def test_save_and_load_checkpoints(self):
        net = Net()
        optim = torch.optim.Adam(params=net.parameters())
        ptvar = torch.tensor(1)
        svar = "test"
        lvar = (1, 2, 3, 4)

        self.experiment.net = net
        self.experiment.optim = optim
        self.experiment.ptvar = ptvar
        self.experiment.svar = svar
        self.experiment.lvar = lvar
        self.experiment.results["test"] = 1

        self.experiment.save_checkpoint(name="test_checkpoint")

        self.experiment.net = Net()
        self.experiment.optim = torch.optim.Adam(params=Net().parameters(), lr=1)
        self.experiment.ptvar = torch.tensor(0)
        self.experiment.svar = "test2"
        self.experiment.lvar = (-1, -2)
        self.experiment.results["test"] = 0

        self.experiment.load_checkpoint(name="test_checkpoint")

        self.assertTrue((list(self.experiment.net.parameters())[0] - list(net.parameters())[0]).sum().item() < 0.00001,
                        "Net could not be restored from checkpoint")
        self.assertTrue(
            self.experiment.optim.state_dict()['param_groups'][0]['lr'] == optim.state_dict()['param_groups'][0]['lr'],
            "optim could not be restored from checkpoint")
        self.assertTrue(self.experiment.ptvar == ptvar.item(), "ptvar could not be restored from checkpoint")
        self.assertTrue(self.experiment.svar == svar, "svar could not be restored from checkpoint")
        self.assertTrue(self.experiment.lvar == lvar, "lvar could not be restored from checkpoint")
        self.assertTrue(self.experiment.results["test"] == 1, "Results could not be restored from checkpoint")

    def test_add_results(self):
        self.experiment.add_result(name="test", value=1)

        self.assertTrue(self.experiment.results["test"] == 1,
                        "Result was not added")
        self.assertTrue(self.experiment.get_result("test") == 1,
                        "Result was not added")

        self.assertTrue(os.path.exists(os.path.join(self.experiment.elog.result_dir, "results-log.json")),
                        "result file could not be stored")

        with open(os.path.join(self.experiment.elog.result_dir, "results-log.json"), "r") as f:
            content = f.read()
        self.assertTrue("test" in content and "1" in content, "results content not sucessfully saved")

    def test_save_tmp_results(self):
        self.experiment.add_result(name="test", value=1)
        self.experiment.run()

        self.assertTrue(os.path.exists(os.path.join(self.experiment.elog.result_dir, "results.json")),
                        "result file could not be stored")

        with open(os.path.join(self.experiment.elog.result_dir, "results.json"), "r") as f:
            content = f.read()
        self.assertTrue("test" in content and "1" in content, "results content not sucessfully temporarily saved")

    def test_save_tmp_checkpoint(self):
        self.experiment.test_var = "test"
        self.experiment.run()

        time.sleep(5)

        self.assertTrue(os.path.exists(os.path.join(self.experiment.elog.checkpoint_dir, "checkpoint_current.pth.tar")),
                        "Temp Checkpoint file could not be stored")
        self.assertTrue(os.path.exists(os.path.join(self.experiment.elog.checkpoint_dir, "checkpoint_last.pth.tar")),
                        "Last Checkpoint file could not be stored")

        self.experiment.test_var = "test2"
        self.experiment.load_checkpoint("checkpoint_current")
        self.assertTrue(self.experiment.test_var == "test",
                        "Temp Checkpoint file loading could not sucessful")
        self.experiment.test_var = "test2"
        self.experiment.load_checkpoint("checkpoint_last")
        self.assertTrue(self.experiment.test_var == "test",
                        "Temp Checkpoint file loading could not sucessful")

    def test_save_exp_info(self):
        self.experiment.run()

        with open(os.path.join(self.experiment.elog.config_dir, "exp.json"), "r") as f:
            exp_info = json.load(f)

        self.assertTrue(exp_info['epoch'] == 10, "Epoch not sucessfully stored in exp info")
        self.assertTrue(exp_info['name'] == 'test_experiment', "Name not sucessfully stored in exp info")
        self.assertTrue(exp_info['state'] == 'Trained', "State not sucessfully stored as 'Trained' in exp info")

    def test_print(self):
        self.experiment.print("0o zD 0o")
        with open(os.path.join(self.experiment.elog.log_dir, "default.log"), "r") as f:
            content = f.read()
        self.assertTrue("0o zD 0o" in content, "Print not sucessfully saved")

    def test_resume(self):
        # TODO

        self.cntr = []
        self.experiment.train = lambda epoch: self.cntr.append(0)
        self.experiment.run()
        self.assertTrue(len(self.cntr) == 10, "Did not call train for the right number of epochs")

        exp2 = PytorchExperiment(name="test-exp2", base_dir=self.test_dir, resume=self.experiment.elog.work_dir)
        exp3 = PytorchExperiment(name="test-exp2", base_dir=self.test_dir, resume=self.experiment.elog.work_dir,
                                 resume_reset_epochs=False)

        exp2.prepare_resume()
        exp3.prepare_resume()

        self.assertTrue(exp2.exp_name == 'test_experiment', "Did not restore exp_name")
        self.assertTrue(exp2._epoch_idx == 0, "Did not reset epochs")
        self.assertTrue(exp3._epoch_idx == 10, "Did reset epochs")

        exp2.train = lambda epoch: self.cntr.append(0)
        exp3.train = lambda epoch: self.cntr.append(0)
        exp2.run()
        self.assertTrue(len(self.cntr) == 20, "Did call not train for exp2")
        exp3.run()
        self.assertTrue(len(self.cntr) == 20, "Did call train for exp3")

        exp2._exp_state = "Ended"
        exp3._exp_state = "Ended"



class Net(nn.Module):
    """
    Small network to test save/load functionality
    """

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


if __name__ == '__main__':
    unittest.main()



import copy
import itertools
import json
import os
import time
import warnings
from collections import defaultdict

import numpy as np

from trixi.logger import ExperimentLogger
from trixi.util import Config
from trixi.util.util import StringMultiTypeDecoder


class ExperimentReader(object):
    """Reader class to read out experiments created by :class:`trixi.experimentlogger.ExperimentLogger`.

    Args:
        work_dir (str): Directory with the structure defined by
                        :class:`trixi.experimentlogger.ExperimentLogger`.
        name (str): Optional name for the experiment. If None, will try
                    to read name from experiment config.

    """

    def __init__(self, base_dir, exp_dir="", name=None, decode_config_clean_str=True):

        super(ExperimentReader, self).__init__()

        self.base_dir = base_dir
        self.exp_dir = exp_dir
        self.work_dir = os.path.abspath(os.path.join(self.base_dir, self.exp_dir))
        self.config_dir = os.path.join(self.work_dir, "config")
        self.log_dir = os.path.join(self.work_dir, "log")
        self.checkpoint_dir = os.path.join(self.work_dir, "checkpoint")
        self.img_dir = os.path.join(self.work_dir, "img")
        self.plot_dir = os.path.join(self.work_dir, "plot")
        self.save_dir = os.path.join(self.work_dir, "save")
        self.result_dir = os.path.join(self.work_dir, "result")

        self.config = Config()
        if decode_config_clean_str:
            self.config.load(os.path.join(self.config_dir, "config.json"), decoder_cls_=StringMultiTypeDecoder)
        else:
            self.config.load(os.path.join(self.config_dir, "config.json"), decoder_cls_=None)

        self.exp_info = Config()
        exp_info_file = os.path.join(self.config_dir, "exp.json")
        if os.path.exists(exp_info_file):
            self.exp_info.load(exp_info_file)

        self.__results_dict = None

        self.meta_name = None
        self.meta_star = False
        self.meta_ignore = False
        self.read_meta_info()

        if name is not None:
            self.exp_name = name
        elif self.meta_name is not None:
            self.exp_name = self.meta_name
        elif "name" in self.exp_info:
            self.exp_name = self.exp_info['name']
        elif "exp_name" in self.config:
            self.exp_name = self.config['exp_name']
        else:
            self.exp_name = "experiments"

        self.ignore = self.meta_ignore
        self.star = self.meta_star

    @staticmethod
    def get_file_contents(folder):
        """Get all files in a folder.

        Returns:
            list: All files joined with folder path.
        """

        if os.path.isdir(folder):
            list_ = map(lambda x: os.path.join(folder, x), sorted(os.listdir(folder)))
            return list(filter(lambda x: os.path.isfile(x), list_))
        else:
            return []

    def get_images(self):
        imgs = []
        imgs += ExperimentReader.get_file_contents(self.img_dir)
        if os.path.isdir(self.img_dir):
            for f in os.listdir(self.img_dir):
                f = os.path.join(self.img_dir, f)
                if os.path.isdir(f):
                    imgs += ExperimentReader.get_file_contents(f)
        return imgs

    def get_plots(self):
        return ExperimentReader.get_file_contents(self.plot_dir)

    def get_checkpoints(self):
        return ExperimentReader.get_file_contents(self.checkpoint_dir)

    def get_logs(self):
        return ExperimentReader.get_file_contents(self.log_dir)

    def get_log_file_content(self, file_name):
        """Read out log file and HTMLify.

        Args:
            file_name (str): Name of the log file.

        Returns:
            str: Log file contents as HTML ready string.
        """

        content = ""
        log_file = os.path.join(self.log_dir, file_name)

        if os.path.exists(log_file):
            with open(log_file, 'r') as f:
                content = f.read()
                content = content.replace("\n", "<br>")

        return content

    def get_results_log(self):
        """Build result dictionary.

        During the experiment result items are
        written out as a stream of quasi-atomic units. This reads the stream and
        builds arrays of corresponding items.
        The resulting dict looks like this::

            {
                "result group": {
                    "result": {
                        "counter": x-array,
                        "data": y-array
                    }
                }
            }

        Returns:
            dict: Result dictionary.

        """

        results_merged = {}

        results = []
        try:
            with open(os.path.join(self.result_dir, "results-log.json"), "r") as results_file:
                results = json.load(results_file)
        except Exception as e:
            try:
                with open(os.path.join(self.result_dir, "results-log.json"), "r") as results_file:
                    results_str = results_file.readlines()
                    if "]" in "]" in "".join(results_str):
                        results_str = [rs.replace("]", ",") for rs in results_str]
                    results_str.append("{}]")
                    results_json = "".join(results_str)
                    results = json.loads(results_json)
            except Exception as ee:
                print("Could not load result log from", self.result_dir)
                print(ee)

        for result in results:
            for key in result.keys():
                counter = result[key]["counter"]
                data = result[key]["data"]
                label = str(result[key]["label"])
                if label not in results_merged:
                    results_merged[label] = {}
                if key not in results_merged[label]:
                    results_merged[label][key] = defaultdict(list)
                results_merged[label][key]["data"].append(data)
                results_merged[label][key]["counter"].append(counter)
                if "max" in result[key] and "min" in result[key]:
                    results_merged[label][key]["max"].append(result[key]["max"])
                    results_merged[label][key]["min"].append(result[key]["min"])

        return results_merged

    def get_results(self):
        """Get the last result item.

        Returns:
            dict: The last result item in the experiment.

        """

        if self.__results_dict is None:

            self.__results_dict = {}
            results_file = os.path.join(self.result_dir, "results.json")

            if os.path.exists(results_file):
                try:
                    with open(results_file, "r") as f:
                        self.__results_dict = json.load(f)
                except Exception as e:
                    pass

        return self.__results_dict

    def ignore_experiment(self):
        """Create a flag file, so the browser ignores this experiment."""
        self.update_meta_info(ignore=True)

    def read_meta_info(self):
        """Reads the meta info of the experiment i.e. new name, stared or ignored"""
        meta_dict = {}
        meta_file = os.path.join(self.work_dir, ".exp_info")
        if os.path.exists(meta_file):
            with open(meta_file, "r") as mf:
                meta_dict = json.load(mf)
            self.meta_name = meta_dict.get("name")
            self.meta_star = meta_dict.get("star", False)
            self.meta_ignore = meta_dict.get("ignore", False)

    def update_meta_info(self, name=None, star=None, ignore=None):
        """
        Updates the meta info i.e. new name, stared or ignored and saves it in the experiment folder

        Args:
            name (str): New name of the experiment
            star (bool): Flag, if experiment is starred/ favorited
            ignore (boll): Flag, if experiment should be ignored
        """

        if name is not None:
            self.meta_name = name
        if star is not None:
            self.meta_star = star
        if ignore is not None:
            self.meta_ignore = ignore

        meta_dict = {
            "name": self.meta_name,
            "star": self.meta_star,
            "ignore": self.meta_ignore
        }
        meta_file = os.path.join(self.work_dir, ".exp_info")
        with open(meta_file, "w") as mf:
            json.dump(meta_dict, mf)


class CombiExperimentReader(ExperimentReader):

    def __init__(self, base_dir, exp_dirs=(), name=None, decode_config_clean_str=True):

        self.base_dir = base_dir

        if name is None or name == "":
            self.exp_name = "combi-experiments"
        else:
            self.exp_name = name

        self.experiments = []
        for exp_dir in exp_dirs:
            self.experiments.append(ExperimentReader(base_dir=base_dir, exp_dir=exp_dir,
                                                     decode_config_clean_str=decode_config_clean_str))

        exp_base_dirs = os.path.commonpath([os.path.dirname(e.work_dir) for e in self.experiments])
        if exp_base_dirs != "":
            self.base_dir = exp_base_dirs

        self.exp_info = Config()

        self.exp_info["epoch"] = -1
        self.exp_info["name"] = self.exp_name
        self.exp_info["state"] = "Combined"
        self.exp_info["time"] = time.strftime("%y-%m-%d_%H:%M:%S", time.localtime(time.time()))

        self.__results_dict = None

        self.work_dir = None
        self.config_dir = "not_saved_yet"
        self.log_dir = "not_saved_yet"
        self.checkpoint_dir = "not_saved_yet"
        self.img_dir = "not_saved_yet"
        self.plot_dir = "not_saved_yet"
        self.save_dir = "not_saved_yet"
        self.result_dir = "not_saved_yet"
        self.exp_dir = "not_saved_yet"

        self.meta_name = None
        self.meta_star = False
        self.meta_ignore = False

        self.config = self.get_config()

        self.elog = None

    def get_config(self):
        combi_config = copy.deepcopy(self.experiments[0].config)
        config_diff = Config.difference_config_static(*[e.config for e in self.experiments], only_set=True)
        combi_config.update(config_diff)

        return combi_config

    def get_results_log(self):
        exp_results_logs = [e.get_results_log() for e in self.experiments]

        log_tags = set()
        for res in exp_results_logs:
            log_tags.update(res.keys())

        res_keys = defaultdict(set)
        for tag in log_tags:
            for exp_res in exp_results_logs:
                if tag not in exp_res:
                    continue
                res_keys[tag].update(exp_res[tag].keys())

        combi_res_log = dict()

        for tag, keys in res_keys.items():
            key_result_dict = {}
            for s_key in keys:
                skey_result = defaultdict(list)

                for exp_res in exp_results_logs:
                    if tag not in exp_res:
                        continue
                    if s_key not in exp_res[tag]:
                        continue

                    for c, val in zip(exp_res[tag][s_key]["counter"], exp_res[tag][s_key]["data"]):
                        skey_result[c].append(val)

                key_result_dict[s_key] = skey_result
            combi_res_log[tag] = key_result_dict

        final_results_log = {}

        for tag, key_result_dict in combi_res_log.items():
            final_results_log[tag] = {}
            for s_key, s_key_result_dict in key_result_dict.items():
                final_results_log[tag][s_key] = defaultdict(list)
                cnts = sorted(s_key_result_dict.keys())
                for cnt in cnts:
                    val_list = s_key_result_dict[cnt]
                    final_results_log[tag][s_key]["counter"].append(cnt)
                    final_results_log[tag][s_key]["data"].append(np.nanmedian(val_list))
                    final_results_log[tag][s_key]["label"].append(tag)
                    final_results_log[tag][s_key]["mean"].append(np.nanmean(val_list))
                    final_results_log[tag][s_key]["median"].append(np.nanmedian(val_list))
                    final_results_log[tag][s_key]["max"].append(np.nanmax(val_list))
                    final_results_log[tag][s_key]["min"].append(np.nanmin(val_list))
                    final_results_log[tag][s_key]["std"].append(np.nanstd(val_list))

        return final_results_log

    def get_results_full(self):
        exp_results = [e.get_results() for e in self.experiments]
        result_keys = set()
        for res in exp_results:
            result_keys.update(res.keys())

        res_collect = defaultdict(list)

        for key in result_keys:
            for exp_res in exp_results:
                if key not in exp_res:
                    continue
                res_collect[key].append(exp_res[key])

        results_dict = {}
        results_aux_dict = defaultdict(dict)

        for key, val_list in res_collect.items():
            results_dict[key] = np.nanmedian(val_list)
            results_aux_dict[key]["mean"] = np.nanmean(val_list)
            results_aux_dict[key]["median"] = np.nanmedian(val_list)
            results_aux_dict[key]["max"] = np.nanmax(val_list)
            results_aux_dict[key]["min"] = np.nanmin(val_list)
            results_aux_dict[key]["std"] = np.nanstd(val_list)
            results_aux_dict[key]["all"] = val_list

        return results_dict, results_aux_dict

    def get_results(self):
        results_dict, _ = self.get_results_full()
        return results_dict

    def get_result_log_dict(self):

        results_dict = self.get_results_log()

        res_list = []

        for tag, key_result_dict in results_dict.items():
            for s_key, s_key_result_dict in key_result_dict.items():
                for cnt, val, max_, min_ in zip(s_key_result_dict["counter"], s_key_result_dict["data"],
                                              s_key_result_dict["max"], s_key_result_dict["min"]):
                    res_list.append({s_key: dict(data=val, counter=cnt, epoch=-1, label=tag, max=max_, min=min_)})

        return res_list

    def ignore_experiment(self):
        """Create a flag file, so the browser ignores this experiment."""
        if self.work_dir is None:
            warnings.warn("Can only be called for a combined experiment which is saved")
            return
        super(CombiExperimentReader, self).ignore_experiment()

    def read_meta_info(self):
        """Reads the meta info of the experiment i.e. new name, stared or ignored"""
        if self.work_dir is None:
            warnings.warn("Can only be called for a combined experiment which is saved")
            return
        super(CombiExperimentReader, self).read_meta_info()

    def update_meta_info(self, name=None, star=None, ignore=None):
        if self.work_dir is None:
            warnings.warn("Can only be called for a combined experiment which is saved")
            return
        super(CombiExperimentReader, self).update_meta_info(name, star, ignore)

    def save(self, target_dir=None):

        if target_dir is None:
            target_dir = self.base_dir

        self.elog = ExperimentLogger(experiment_name=self.exp_name, base_dir=target_dir)

        self.exp_dir = self.elog.folder_name
        self.work_dir = self.elog.work_dir
        self.config_dir = os.path.join(self.work_dir, "config")
        self.log_dir = os.path.join(self.work_dir, "log")
        self.checkpoint_dir = os.path.join(self.work_dir, "checkpoint")
        self.img_dir = os.path.join(self.work_dir, "img")
        self.plot_dir = os.path.join(self.work_dir, "plot")
        self.save_dir = os.path.join(self.work_dir, "save")
        self.result_dir = os.path.join(self.work_dir, "result")

        def __default(o):
            if isinstance(o, np.int64): return int(o)
            raise TypeError

        results, results_combi = self.get_results_full()
        self.config.dump(os.path.join(self.elog.config_dir, "config.json"))
        self.elog.save_config(self.exp_info, "exp")
        self.elog.save_result(results, "results", encoder_cls=None, default=__default)
        self.elog.save_result(results_combi, "results-combi", encoder_cls=None, default=__default)
        self.elog.save_result(self.get_result_log_dict(), "results-log", encoder_cls=None, default=__default)
        self.elog.text_logger.log_to("\n".join(["\n"]+[e.work_dir for e in self.experiments]), "combi_exps")



def group_experiments_by(exps, group_by_list):
    configs_flat = [e.config.flat() for e in exps]  ### Exclude already combined experiments
    config_diff = Config.difference_config_static(*configs_flat)

    group_diff_key_list = []
    group_diff_val_list = []

    for diff_key in group_by_list:
        if diff_key in config_diff:
            group_diff_key_list.append(diff_key)
            group_diff_val_list.append(set(config_diff[diff_key]))

    val_combis = itertools.product(*group_diff_val_list)

    group_dict = defaultdict(list)

    for val_combi in val_combis:
        for e in exps:
            e_config = e.config.flat()
            is_match = True

            for key, val in zip(group_diff_key_list, val_combi):
                if e_config[key] != val:
                    is_match = False

            if is_match:
                group_dict[val_combi].append(e)

    return list(group_dict.values())



from trixi.experiment.experiment import Experiment

try:
    from trixi.experiment.pytorchexperiment import PytorchExperiment
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))



from trixi.util.util import (
    CustomJSONEncoder,
    CustomJSONDecoder,
    MultiTypeEncoder,
    MultiTypeDecoder,
    ModuleMultiTypeEncoder,
    ModuleMultiTypeDecoder,
    Singleton,
    savefig_and_close,
    random_string,
    create_folder,
    name_and_iter_to_filename,
    SafeDict,
    PyLock,
    LogDict,
    ResultLogDict,
    ResultElement
)
from trixi.util.config import Config
from trixi.util.extravisdom import ExtraVisdom
from trixi.util.sourcepacker import SourcePacker
from trixi.util.gridsearch import GridSearch



import json
from trixi.util import ModuleMultiTypeDecoder


class GridSearch(dict):

    def all_combinations(self):

        if len(self) == 0:
            return []

        combinations = []
        zero_key = sorted(self.keys())[0]
        inner_dict = self.copy()
        del inner_dict[zero_key]

        for val in self[zero_key]:

            if len(inner_dict) == 0:
                combinations.append({zero_key: val})
            else:
                for combination in GridSearch(inner_dict):
                    combination[zero_key] = val
                    combinations.append(combination)

        return combinations

    def __iter__(self):

        self._counter = 0
        self._all_combinations = self.all_combinations()
        self._max = len(self._all_combinations)
        return self

    def __next__(self):

        if self._counter >= self._max:
            raise StopIteration

        self._counter += 1
        return self._all_combinations[self._counter - 1]

    def read(self, file_, raise_=True, decoder_cls_=ModuleMultiTypeDecoder, **kwargs):

        try:
            if hasattr(file_, "read"):
                new_dict = json.load(file_, cls=decoder_cls_, **kwargs)
            else:
                with open(file_, "r") as file_object:
                    new_dict = json.load(file_object, cls=decoder_cls_, **kwargs)
        except Exception as e:
            if raise_:
                raise e

        self.update(new_dict)

        return self



from __future__ import print_function

import datetime
import logging
import os
import sys

from trixi.logger.abstractlogger import AbstractLogger
from trixi.util import random_string


class TextFileLogger(AbstractLogger):
    """
    A Logger for logging text into different text files and output streams (using the python logging framework)
    """

    def __init__(self, base_dir=None, logging_level=logging.DEBUG, logging_stream=sys.stdout,
                 default_stream_handler=True, **kwargs):
        """
        Initializes a TextFileLogger and a default logger

        Args:
            base_dir: Directory to save the log files in
            logging_level: Default logging level for the default log
            logging_stream: default logging level
            default_stream_handler: Falg, if a default stream handler should be added
        """

        super(TextFileLogger, self).__init__(**kwargs)

        self.base_dir = base_dir
        self.logging_level = logging_level
        self.logging_stream = logging_stream
        self.loggers = dict()
        self.stream_handler_formatter = logging.Formatter('%(message)s')
        self.file_handler_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

        self.init_time = datetime.datetime.today()

        # set up logging
        self.logging_identifier = random_string(10)
        self.add_logger("default", stream_handler=default_stream_handler)

    def add_logger(self, name, logging_level=None, file_handler=True, stream_handler=True):
        """
        Adds a new logger

        Args:
            name: Name of the new logger
            logging_level: Logging level of the new logger
            file_handler: Flag, if it should use a file_handler, if yes creates a new file with the given name in the
                logging directory
            stream_handler: Flag, if the logger should also log to the default stream

        Returns:

        """

        self.loggers[name] = logging.getLogger(name + "-" + self.logging_identifier)

        if logging_level is None:
            self.loggers[name].setLevel(self.logging_level)
        else:
            self.loggers[name].setLevel(logging_level)

        if file_handler is True:
            self.add_file_handler(name, name)
        elif isinstance(file_handler, (list, tuple, set)):
            for fh in file_handler:
                if isinstance(fh, logging.FileHandler):
                    self.add_handler(fh, name)
                else:
                    self.add_file_handler(fh, name)
        elif isinstance(file_handler, logging.FileHandler):
            self.add_handler(file_handler, name)
        else:
            pass

        if stream_handler:
            self.add_stream_handler(name)

    def add_handler(self, handler, logger="default"):
        """
        Adds an additional handler to a logger with a name
        Args:
            handler: Logging handler to be added to a given logger
            logger: Name of the logger to add the hander to

        """
        self.loggers[logger].addHandler(handler)

    def add_file_handler(self, name, logger="default"):
        """
        Adds a file handler to a logger, thus the logger will log into a log file with a given name

        Args:
            name: File name of the log file (in which the logger will log now)
            logger: Name of the logger to add the file-hander/ logging file to

        """
        file_handler = logging.FileHandler(os.path.join(self.base_dir, name + ".log"))
        file_handler.setFormatter(self.file_handler_formatter)
        self.add_handler(file_handler, logger)

    def add_stream_handler(self, logger="default"):
        """
        Adds a stream handler to a logger, thus the logger will log into the default logging stream

        Args:
            logger: Name of the logger to add the stream-hander to

        """
        stream_handler = logging.StreamHandler(self.logging_stream)
        stream_handler.setFormatter(self.stream_handler_formatter)
        self.add_handler(stream_handler, logger)

    def print(self, *args, logger="default"):
        """
        Prints and logs objects

        Args:
            *args: Object to print/ log
            logger: Logger which should log

        """
        self.log(" ".join(map(str, args)), logger)

    def log(self, msg, logger="default"):
        """
        Prints and logs a message with the level info

        Args:
            msg: Message to print/ log
            logger: Logger which should log

        """
        self.loggers[logger].info(msg)

    def info(self, msg, logger="default"):
        """
        Prints and logs a message with the level info

        Args:
            msg: Message to print/ log
            logger: Logger which should log

        """
        self.loggers[logger].info(msg)

    def debug(self, msg, logger="default"):
        """
        Prints and logs a message with the level debug

        Args:
            msg: Message to print/ log
            logger: Logger which should log

        """
        self.loggers[logger].debug(msg)

    def error(self, msg, logger="default"):
        """
        Prints and logs a message with the level error

        Args:
            msg: Message to print/ log
            logger: Logger which should log

        """
        self.loggers[logger].error(msg)

    def log_to(self, msg, name, log_to_default=False):
        """
        Logs to an existing logger or if logger does not exists creates new one

        Args:
            msg: Message to be logged
            name: Name of the logger to log to (usually also the logfile-name)
            log_to_default: Flag if it should in addition to the logger given by name, log to the default logger

        """

        if name not in self.loggers:
            self.add_logger(name)
        self.log(msg, name)
        if log_to_default:
            self.log(msg, "default")

    def show_text(self, text, name=None, logger="default", **kwargs):
        """
        Logs a text. Calls the log function (for combatibility reasons with AbstractLogger)

        Args:
            text: Text to be logged
            name: Some identifier for the text (will be added infront of the text)
            logger: Name of the Logger to log to

        """
        if name is not None:
            self.log("{}: {}".format(name, text), logger)
        else:
            self.log(text, logger)

    def show_value(self, value, name=None, logger="default", **kwargs):
        """
        Logs a Value. Calls the log function (for combatibility reasons with AbstractLogger)

        Args:
            value: Value to be logged
            name: Some identifier for the text (will be added infront of the text)
            logger: Name of the Logger to log to

        """
        if name is not None:
            self.log("{}: {}".format(name, value), logger)
        else:
            self.log(value, logger)



from abc import ABCMeta, abstractmethod
import _thread
from functools import wraps


def convert_params(f):
    """Decorator to call the process_params method of the class."""

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return self.process_params(f, *args, **kwargs)

    return wrapper


def threaded(f):
    """Decorator to run the process in an extra thread."""

    def wrapper(*args, **kwargs):
        return _thread.start_new(f, args, kwargs)

    return wrapper


class AbstractLogger(object):
    """Abstract interface for visual logger."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    def process_params(self, f, *args, **kwargs):
        """
        Implement this to handle data conversions in your logger.

        Example: Implement logger for numpy data, then implement torch logger as child of numpy logger and just use
        the process_params method to convert from torch to numpy.
        """

        return f(self, *args, **kwargs)

    @abstractmethod
    @convert_params
    def show_image(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store an image"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_value(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a value"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_text(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a text"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_barplot(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a barplot"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_lineplot(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a lineplot"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_scatterplot(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a scatterplot"""
        raise NotImplementedError()

    @abstractmethod
    @convert_params
    def show_piechart(self, *args, **kwargs):
        """Abstract method which should handle and somehow log/ store a piechart"""
        raise NotImplementedError()



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import seaborn
from trixi.logger.plt.numpyseabornplotlogger import NumpySeabornPlotLogger

from trixi.logger.abstractlogger import AbstractLogger, convert_params
from trixi.util.util import chw_to_hwc, figure_to_image


class NumpySeabornImagePlotLogger(NumpySeabornPlotLogger):
    """
    Wrapper around :class:`.NumpySeabornPlotLogger` that renders figures into numpy arrays.
    """

    def __init__(self, **kwargs):
        """
        Initializes a new NumpySeabornPlotLogger
        """
        super(NumpySeabornImagePlotLogger, self).__init__(**kwargs)

    @convert_params
    def show_image(self, image, name=None, *args, **kwargs):
        """
        Create an image figure

        Args:
            image: The image array to be displayed
            name: The name of the image window

        Returns:
            A numpy array image of the figure

        """
        # figure = super().show_image(image, name, show=False, *args, **kwargs)
        # figure_image = figure_to_image(figure)

        return image

    @convert_params
    def show_value(self, value, name, counter=None, tag=None, *args, **kwargs):
        """
       Creates a line plot that is automatically appended with new values and returns it as a figure.

       Args:
           value: Value to be plotted / appended to the graph (y-axis value)
           name: The name of the window
           counter: counter, which tells the number of the sample (with the same name) (x-axis value)
           tag: Tag, grouping similar values. Values with the same tag will be plotted in the same plot

        Returns:
            A numpy array image of the figure

        """
        figure = super().show_value(value, name, show=False, counter=counter, tag=tag, *args, **kwargs)
        figure_image = figure_to_image(figure)

        return figure_image

    @convert_params
    def show_barplot(self, array, name=None, *args, **kwargs):
        """
        Creates a bar plot figure from an array

        Args:
            array: array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name: The name of the figure

        Returns:
            A numpy array image of the figure
        """

        figure = super().show_barplot(array, name, show=False, *args, **kwargs)
        figure_image = figure_to_image(figure)

        return figure_image

    @convert_params
    def show_lineplot(self, y_vals, x_vals=None, name=None, *args, **kwargs):
        """
        Creates a line plot figure with (multiple) lines plot, given values Y (and optional the corresponding X values)

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
            not set the points are assumed to be equally distributed in the interval [0, 1] )
            name: The name of the figure

        Returns:
            A numpy array image of the figure
        """

        figure = super().show_lineplot(y_vals=y_vals, x_vals=x_vals, name=name, show=False, *args, **kwargs)
        figure_image = figure_to_image(figure)

        return figure_image

    @convert_params
    def show_scatterplot(self, array, name=None, *args, **kwargs):
        """
        Creates a scatter plot figure with the points given in array

        Args:
            array: A 2d array with size N x dim, where each element i \in N at X[i] results in a a 2d (if dim = 2)/
            3d (if dim = 3) point.
            name: The name of the figure

         Returns:
            A numpy array image of the figure
        """

        figure = super().show_scatterplot(array, name, show=False, *args, **kwargs)
        figure_image = figure_to_image(figure)

        return figure_image

    @convert_params
    def show_piechart(self, array, name=None, *args, **kwargs):
        """
        Creates a scatter plot figure

        Args:
            array: Array of positive integers. Each integer will be presented as a part of the pie (with the total
            as the sum of all integers)
            name: The name of the figure

        Returns:
            A numpy array image of the figure
        """

        figure = super().show_piechart(array, name, show=False, *args, **kwargs)
        figure_image = figure_to_image(figure)

        return figure_image



import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy import misc

from trixi.logger.experiment.experimentlogger import ExperimentLogger
from trixi.util.config import Config

test_dir = "test_dir"


class TestExperimentLogger(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.gettempdir()
        self.experimentLogger = ExperimentLogger(exp_name="test",
                                                 base_dir=self.test_dir,
                                                 folder_format="{experiment_name}")

    def tearDown(self):
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_create_folders(self):
        exp_dir = os.path.join(self.test_dir, "test")
        self.assertTrue(os.path.exists(exp_dir), "Experiment dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "checkpoint")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "config")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "img")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "log")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "config")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "plot")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "result")), "Config dir not created")
        self.assertTrue(os.path.exists(os.path.join(exp_dir, "save")), "Config dir not created")

    def test_two_experiment_loggers_same_test_dir_no_run_number_throws_error(self):
        logger = ExperimentLogger(exp_name="test",
                                  base_dir=self.test_dir,
                                  folder_format="{experiment_name}")
        self.assertTrue(os.path.isdir(logger.base_dir), "Experiment directory not created")

    def test_show_image(self):
        image = np.random.random_sample((3, 128, 128))
        self.experimentLogger.show_image(image, "image")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.img_dir, "image.png")),
                        "Show image could not create image")

    def test_show_barplot(self):
        array = np.random.random_sample(5)
        self.experimentLogger.show_barplot(array, "barplot")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "barplot.png")),
                        "Show barplot could not create barplot")

    def test_show_lineplot(self):
        x = [0, 1, 2, 3, 4, 5]
        y = np.random.random_sample(6)
        self.experimentLogger.show_lineplot(y, x, name="lineplot1")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "lineplot1.png")),
                        "Show lineplot could not create lineplot")
        self.experimentLogger.show_lineplot(y, name="lineplot2")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "lineplot2.png")),
                        "Show lineplot could not create lineplot without x vals")

    def test_show_piechart(self):
        array = np.random.random_sample(5)
        self.experimentLogger.show_piechart(array, "piechart")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "piechart.png")),
                        "Show piechart could not create piechart")

    def test_show_scatterplot(self):
        array = np.random.random_sample((5, 2))
        self.experimentLogger.show_scatterplot(array, "scatterplot")
        self.assertTrue(os.path.exists(os.path.join(self.experimentLogger.plot_dir, "scatterplot.png")),
                        "Show scatterplot could not create scatterplot")

    def test_show_value(self):
        val = np.random.random_sample(1)
        self.experimentLogger.show_value(val, "value")
        plt1_content = misc.imread(os.path.join(self.experimentLogger.plot_dir, "value.png"))

        val = np.random.random_sample(1)
        self.experimentLogger.show_value(val, "value")
        plt2_content = misc.imread(os.path.join(self.experimentLogger.plot_dir, "value.png"))
        self.assertFalse(np.array_equal(plt1_content, plt2_content), "Show value did not update the plot")

        val = np.random.random_sample(1)
        self.experimentLogger.show_value(val, "value", counter=4)
        plt3_content = misc.imread(os.path.join(self.experimentLogger.plot_dir, "value.png"))
        self.assertFalse(np.array_equal(plt2_content, plt3_content), "Show value did not update the plot")

    def test_show_text(self):
        text = "\nTest 4 fun: zD ;-D 0o"
        self.experimentLogger.show_text(text)
        log_text = ""
        with open(os.path.join(self.experimentLogger.log_dir, "default.log"), 'r') as log_file:
            log_text = log_file.read()
        self.assertTrue(text in log_text)



    def test_save_and_load_config(self):
        c = Config()
        c.text = "0o"
        c.nmbr = 4
        c.cls = str
        c.lst = [1, 2, 3, 4]

        self.experimentLogger.save_config(c, "config")
        config_path = os.path.join(self.experimentLogger.config_dir, "config.json")
        self.assertTrue(os.path.exists(config_path),
                        "Config could not be saved")

        c2 = self.experimentLogger.load_config("config")
        self.assertTrue("text" in c2 and c2.text == c.text, "Text in config could not be restored")
        self.assertTrue("nmbr" in c2 and c2.nmbr == c.nmbr, "Number in config could not be restored")
        self.assertTrue("cls" in c2 and c2.cls == c.cls, "Text in config could not be restored")
        self.assertTrue("lst" in c2 and c2.lst == c.lst, "List in config could not be restored")

    def test_save_results(self):
        d = dict()
        d['best_val'] = 3.1415926
        d['worst_val'] = 2.718281

        self.experimentLogger.save_result(d, "result")
        result_path = os.path.join(self.experimentLogger.result_dir, "result.json")
        self.assertTrue(os.path.exists(result_path),
                        "Result could not be saved")
        result_text = ""
        with open(result_path, 'r') as log_file:
            result_text = log_file.read()
        self.assertTrue("best_val" in result_text, "Saved vals could not be found in result")
        self.assertTrue("3.1415926" in result_text, "Saved vals could not be found in result")
        self.assertTrue("worst_val" in result_text, "Saved vals could not be found in result")
        self.assertTrue("2.718281" in result_text, "Saved vals could not be found in result")

    def test_save_and_load_dict(self):
        d = dict()
        d['text'] = "0o"
        d['nmbr'] = 4
        d['tuple'] = (1, 2, 3, 4)

        self.experimentLogger.save_dict(d, "dict")
        dict_path = os.path.join(self.experimentLogger.save_dir, "dict.json")
        self.assertTrue(os.path.exists(dict_path),
                        "dict could not be saved")

        d2 = self.experimentLogger.load_dict("dict")
        self.assertTrue("text" in d2 and d2['text'] == d['text'], "Text in dict could not be restored")
        self.assertTrue("nmbr" in d2 and d2['nmbr'] == d['nmbr'], "Number in dict could not be restored")
        self.assertTrue("tuple" in d2 and d2['tuple'] == d['tuple'], "Tuple in dict could not be restored")


    def test_save_and_load_numpy_data(self):
        np_array = np.random.random_sample((3, 128, 128))

        self.experimentLogger.save_numpy_data(np_array, "array")
        np_path = os.path.join(self.experimentLogger.save_dir, "array.npy")
        self.assertTrue(os.path.exists(np_path),
                        "Np Array could not be saved")

        array2 = self.experimentLogger.load_numpy_data("array")
        self.assertTrue(np.array_equal(np_array, array2), "Numpy array could not be loaded")

    def test_save_and_load_pickle(self):
        data1 = np.random.random_sample((3, 128, 128))

        self.experimentLogger.save_pickle(data1, "pickle.pkl")
        np_path = os.path.join(self.experimentLogger.save_dir, "pickle.pkl")
        self.assertTrue(os.path.exists(np_path),
                        "Pickle file could not be saved")

        data2 = self.experimentLogger.load_pickle("pickle.pkl")
        self.assertTrue(np.array_equal(data1, data2), "Pickle file could not be loaded")


if __name__ == '__main__':
    unittest.main()



import atexit
import fnmatch
import json
import os
import numpy as np
import sys
import re
import random
import shutil
import string
import time
import traceback
import warnings

import torch

from trixi.experiment.experiment import Experiment
from trixi.logger import CombinedLogger, PytorchExperimentLogger, PytorchVisdomLogger, TelegramMessageLogger
from trixi.logger.message.slackmessagelogger import SlackMessageLogger
from trixi.logger.tensorboard.pytorchtensorboardxlogger import PytorchTensorboardXLogger
from trixi.util import Config, ResultElement, ResultLogDict, SourcePacker, name_and_iter_to_filename
from trixi.util.config import update_from_sys_argv
from trixi.util.pytorchutils import set_seed

logger_lookup_dict = dict(
    visdom=PytorchVisdomLogger,
    tensorboard=PytorchTensorboardXLogger,
    telegram=TelegramMessageLogger,
    slack=SlackMessageLogger,
)


class PytorchExperiment(Experiment):
    """
    A PytorchExperiment extends the basic
    functionality of the :class:`.Experiment` class with
    convenience features for PyTorch (and general logging) such as creating a folder structure,
    saving, plotting results and checkpointing your experiment.

    The basic life cycle of a PytorchExperiment is the same as
    :class:`.Experiment`::

        setup()
        prepare()

        for epoch in n_epochs:
            train()
            validate()

        end()

    where the distinction between the first two is that between them
    PytorchExperiment will automatically restore checkpoints and save the
    :attr:`_config_raw` in :meth:`._setup_internal`. Please see below for more
    information on this.
    To get your own experiment simply inherit from the PytorchExperiment and
    overwrite the :meth:`.setup`, :meth:`.prepare`, :meth:`.train`,
    :meth:`.validate` method (or you can use the `very` experimental decorator
    :func:`.experimentify` to convert your class into a experiment).
    Then you can run your own experiment by calling the :meth:`.run` method.

    Internally PytorchExperiment will provide a number of member variables which
    you can access.

        - n_epochs
            Number of epochs.
        - exp_name
            Name of your experiment.
        - config
            The (initialized) :class:`.Config` of your experiment. You can
            access the uninitialized one via :attr:`_config_raw`.
        - result
            A dict in which you can store your result values. If a
            :class:`.PytorchExperimentLogger` is used, results will be a
            :class:`.ResultLogDict` that directly automatically writes to a file
            and also stores the N last entries for each key for quick access
            (e.g. to quickly get the running mean).
        - elog (if base_dir is given)
            A :class:`.PytorchExperimentLogger` instance which can log your
            results to a given folder. Will automatically be created if a base_dir is available.
        - loggers
            Contains all loggers you provide, including the experiment logger, accessible by the
            names you provide.
        - clog
            A :class:`.CombinedLogger` instance which logs to all loggers with
            different frequencies (specified with the last entry in the tuple you provide for each
            logger where 1 means every time and N means every Nth time,
            e.g. if you only want to send stuff to Visdom every 10th time).

    The most important attribute is certainly :attr:`.config`, which is the
    initialized :class:`.Config` for the experiment. To understand how it needs
    to be structured to allow for automatic instantiation of types, please refer
    to its documentation. If you decide not to use this functionality,
    :attr:`config` and :attr:`_config_raw` are identical. **Beware however that
    by default the Pytorchexperiment only saves the raw config** after
    :meth:`.setup`. If you modify :attr:`config` during setup, make sure
    to implement :meth:`._setup_internal` yourself should you want the modified
    config to be saved::

        def _setup_internal(self):

            super(YourExperiment, self)._setup_internal() # calls .prepare_resume()
            self.elog.save_config(self.config, "config")

    Args:
        config (dict or Config): Configures your experiment. If :attr:`name`,
            :attr:`n_epochs`, :attr:`seed`, :attr:`base_dir` are given in the
            config, it will automatically
            overwrite the other args/kwargs with the values from the config.
            In addition (defined by :attr:`parse_config_sys_argv`) the config
            automatically parses the argv arguments and updates its values if a
            key matches a console argument.
        name (str):
            The name of the PytorchExperiment.
        n_epochs (int): The number of epochs (number of times the training
            cycle will be executed).
        seed (int): A random seed (which will set the random, numpy and
            torch seed).
        base_dir (str): A base directory in which the experiment result folder
            will be created. A :class:`.PytorchExperimentLogger` instance will be created if
            this is given.
        globs: The :func:`globals` of the script which is run. This is necessary
            to get and save the executed files in the experiment folder.
        resume (str or PytorchExperiment): Another PytorchExperiment or path to
            the result dir from another PytorchExperiment from which it will
            load the PyTorch modules and other member variables and resume
            the experiment.
        ignore_resume_config (bool): If :obj:`True` it will not resume with the
            config from the resume experiment but take the current/own config.
        resume_save_types (list or tuple): A list which can define which values
            to restore when resuming. Choices are:

                - "model" <-- Pytorch models
                - "optimizer" <-- Optimizers
                - "simple" <-- Simple python variables (basic types and lists/tuples
                - "th_vars" <-- torch tensors/variables
                - "results" <-- The result dict

        resume_reset_epochs (bool): Set epoch to zero if you resume an existing experiment.
        parse_sys_argv (bool): Parsing the console arguments (argv) to get a
            :attr:`config path` and/or :attr:`resume_path`.
        parse_config_sys_argv (bool): Parse argv to update the config
            (if the keys match).
        checkpoint_to_cpu (bool): When checkpointing, transfer all tensors to
            the CPU beforehand.
        save_checkpoint_every_epoch (int): Determines after how many epochs a
            checkpoint is stored.
        explogger_kwargs (dict): Keyword arguments for :attr:`elog`
            instantiation.
        explogger_freq (int): The frequency x (meaning one in x) with which
            the :attr:`clog` will call the :attr:`elog`.
        loggers (dict): Specify additional loggers.
            Entries should have one of these formats::

                "name": "identifier" (will default to a frequency of 10)
                "name": ("identifier"(, kwargs, frequency)) (last two are optional)

            "identifier" is one of "telegram", "tensorboard", "visdom", "slack".
        append_rnd_to_name (bool): If :obj:`True`, will append a random six
            digit string to the experiment name.

     """

    def __init__(self,
                 config=None,
                 name=None,
                 n_epochs=None,
                 seed=None,
                 base_dir=None,
                 globs=None,
                 resume=None,
                 ignore_resume_config=False,
                 resume_save_types=("model", "optimizer", "simple", "th_vars", "results"),
                 resume_reset_epochs=True,
                 parse_sys_argv=False,
                 checkpoint_to_cpu=True,
                 save_checkpoint_every_epoch=1,
                 explogger_kwargs=None,
                 explogger_freq=1,
                 loggers=None,
                 append_rnd_to_name=False):

        # super(PytorchExperiment, self).__init__()
        Experiment.__init__(self)

        # check for command line inputs for config_path and resume_path,
        # will be prioritized over config and resume!
        config_path_from_argv = None
        if parse_sys_argv:
            config_path_from_argv, resume_path_from_argv = get_vars_from_sys_argv()
            if resume_path_from_argv:
                resume = resume_path_from_argv

        # construct _config_raw
        if config_path_from_argv is None:
            self._config_raw = self._config_raw_from_input(config, name, n_epochs, seed, append_rnd_to_name)
        else:
            self._config_raw = Config(file_=config_path_from_argv)
        update_from_sys_argv(self._config_raw)

        # set a few experiment attributes
        self.n_epochs = self._config_raw["n_epochs"]
        self._seed = self._config_raw['seed']
        set_seed(self._seed)
        self.exp_name = self._config_raw["name"]
        self._checkpoint_to_cpu = checkpoint_to_cpu
        self._save_checkpoint_every_epoch = save_checkpoint_every_epoch
        self.results = dict()

        # get base_dir from _config_raw or store there
        if base_dir is not None:
            self._config_raw["base_dir"] = base_dir
        base_dir = self._config_raw["base_dir"]

        # Construct experiment logger (automatically activated if base_dir is there)
        self.loggers = {}
        logger_list = []
        if base_dir is not None:
            if explogger_kwargs is None:
                explogger_kwargs = {}
            self.elog = PytorchExperimentLogger(base_dir=base_dir,
                                                exp_name=self.exp_name,
                                                **explogger_kwargs)
            if explogger_freq is not None and explogger_freq > 0:
                logger_list.append((self.elog, explogger_freq))
            self.results = ResultLogDict("results-log.json", base_dir=self.elog.result_dir)
        else:
            self.elog = None

        # Construct other loggers
        if loggers is not None:
            for logger_name, logger_cfg in loggers.items():
                _logger, log_freq = self._make_logger(logger_name, logger_cfg)
                self.loggers[logger_name] = _logger
                if log_freq is not None and log_freq > 0:
                    logger_list.append((_logger, log_freq))

        self.clog = CombinedLogger(*logger_list)

        # Set resume attributes and update _config_raw,
        # actual resuming is done automatically after setup in _setup_internal
        self._resume_path = None
        self._resume_save_types = resume_save_types
        self._ignore_resume_config = ignore_resume_config
        self._resume_reset_epochs = resume_reset_epochs
        if resume is not None:
            if isinstance(resume, str):
                if resume == "last":
                    if base_dir is None:
                        raise ValueError("resume='last' requires base_dir.")
                    self._resume_path = os.path.join(base_dir, sorted(os.listdir(base_dir))[-1])
                else:
                    self._resume_path = resume
            elif isinstance(resume, PytorchExperiment):
                self._resume_path = resume.elog.base_dir
        if self._resume_path is not None and not self._ignore_resume_config:
            self._config_raw.update(Config(file_=os.path.join(self._resume_path, "config", "config.json")),
                                    ignore=list(map(lambda x: re.sub("^-+", "", x), sys.argv)))

        # Save everything we need to reproduce experiment
        if globs is not None and self.elog is not None:
            zip_name = os.path.join(self.elog.save_dir, "sources.zip")
            SourcePacker.zip_sources(globs, zip_name)

        # Init objects in config
        self.config = Config.init_objects(self._config_raw)

        atexit.register(self.at_exit_func)

    def _config_raw_from_input(self,
                               config=None,
                               name=None,
                               n_epochs=None,
                               seed=None,
                               append_rnd_to_name=False):
        """Construct _config_raw from input."""

        _config_raw = None
        if isinstance(config, str):
            _config_raw = Config(file_=config)
        elif isinstance(config, (Config, dict)):
            _config_raw = Config(config=config)
        else:
            _config_raw = Config()

        if n_epochs is None and _config_raw.get("n_epochs") is not None:
            n_epochs = _config_raw["n_epochs"]
        elif n_epochs is None and _config_raw.get("n_epochs") is None:
            n_epochs = 0
        _config_raw["n_epochs"] = n_epochs

        if seed is None and _config_raw.get('seed') is not None:
            seed = _config_raw['seed']
        elif seed is None and _config_raw.get('seed') is None:
            random_data = os.urandom(4)
            seed = int.from_bytes(random_data, byteorder="big")
        _config_raw['seed'] = seed

        if name is None and _config_raw.get("name") is not None:
            name = _config_raw["name"]
        elif name is None and _config_raw.get("name") is None:
            name = "experiment"
        if append_rnd_to_name:
            rnd_str = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(5))
            name += "_" + rnd_str
        _config_raw["name"] = name

        return _config_raw

    def _make_logger(self, logger_name, logger_cfg):

        if isinstance(logger_cfg, (list, tuple)):
            log_type = logger_cfg[0]
            log_params = logger_cfg[1] if len(logger_cfg) > 1 else {}
            log_freq = logger_cfg[2] if len(logger_cfg) > 2 else 10
        else:
            assert isinstance(logger_cfg, str), "The specified logger has to either be a string or a list with " \
                                                "name, parameters, clog_frequency"
            log_type = logger_cfg
            log_params = {}
            log_freq = 10

        if "exp_name" not in log_params:
            log_params["exp_name"] = self.exp_name

        if log_type == "tensorboard":
            if "target_dir" not in log_params or log_params["target_dir"] is None:
                if self.elog is not None:
                    log_params["target_dir"] = os.path.join(self.elog.save_dir, "tensorboard")
                else:
                    raise AttributeError("TensorboardLogger requires a target_dir or an ExperimentLogger instance.")
            elif self.elog is not None:
                log_params["target_dir"] = os.path.join(log_params["target_dir"], self.elog.folder_name)

        log_type = logger_lookup_dict[log_type]
        _logger = log_type(**log_params)

        return _logger, log_freq

    @property
    def vlog(self):
        if "visdom" in self.loggers:
            return self.loggers["visdom"]
        elif "v" in self.loggers:
            return self.loggers["v"]
        else:
            return None

    @property
    def tlog(self):
        if "telegram" in self.loggers:
            return self.loggers["telegram"]
        elif "t" in self.loggers:
            return self.loggers["t"]
        else:
            return None

    @property
    def txlog(self):
        if "tensorboard" in self.loggers:
            return self.loggers["tensorboard"]
        if "tensorboardx" in self.loggers:
            return self.loggers["tensorboardx"]
        elif "tx" in self.loggers:
            return self.loggers["tx"]
        else:
            return None

    @property
    def slog(self):
        if "slack" in self.loggers:
            return self.loggers["slack"]
        elif "s" in self.loggers:
            return self.loggers["s"]
        else:
            return None

    def process_err(self, e):
        if self.elog is not None:
            self.elog.text_logger.log_to("\n".join(traceback.format_tb(e.__traceback__)), "err")

    def update_attributes(self, var_dict, ignore=()):
        """
        Updates the member attributes with the attributes given in the var_dict

        Args:
            var_dict (dict): dict in which the update values stored. If a key matches a member attribute name
                the member attribute will be updated
            ignore (list or tuple): iterable of keys to ignore

        """
        for key, val in var_dict.items():
            if key == "results":
                self.results.load(val)
                continue
            if key in ignore:
                continue
            if hasattr(self, key):
                setattr(self, key, val)

    def get_pytorch_modules(self, from_config=True):
        """
        Returns all torch.nn.Modules stored in the experiment in a dict (even child dicts are stored).

        Args:
            from_config (bool): Also get modules that are stored in the :attr:`.config` attribute.

        Returns:
            dict: Dictionary of PyTorch modules

        """

        def parse_torchmodules_recursive(input, output):
            if isinstance(input, dict):
                for key, value in input.items():
                    if isinstance(value, dict):
                        parse_torchmodules_recursive(value, output)
                    elif isinstance(value, torch.nn.Module):
                        output[key] = value

        pyth_modules = dict()
        parse_torchmodules_recursive(self.__dict__, pyth_modules)
        # for key, val in self.__dict__.items():
        #     if isinstance(val, torch.nn.Module):
        #         pyth_modules[key] = val
        if from_config:
            for key, val in self.config.items():
                if isinstance(val, torch.nn.Module):
                    if type(key) == str:
                        key = "config." + key
                    pyth_modules[key] = val
        return pyth_modules

    def get_pytorch_optimizers(self, from_config=True):
        """
        Returns all torch.optim.Optimizers stored in the experiment in a dict.

        Args:
            from_config (bool): Also get optimizers that are stored in the :attr:`.config`
                attribute.

        Returns:
            dict: Dictionary of PyTorch optimizers

        """

        pyth_optimizers = dict()
        for key, val in self.__dict__.items():
            if isinstance(val, torch.optim.Optimizer):
                pyth_optimizers[key] = val
        if from_config:
            for key, val in self.config.items():
                if isinstance(val, torch.optim.Optimizer):
                    if type(key) == str:
                        key = "config." + key
                    pyth_optimizers[key] = val
        return pyth_optimizers

    def get_simple_variables(self, ignore=()):
        """
        Returns all standard variables in the experiment in a dict.
        Specifically, this looks for types :class:`int`, :class:`float`, :class:`bytes`,
        :class:`bool`, :class:`str`, :class:`set`, :class:`list`, :class:`tuple`.

        Args:
            ignore (list or tuple): Iterable of names which will be ignored

        Returns:
            dict: Dictionary of variables

        """

        simple_vars = dict()
        for key, val in self.__dict__.items():
            if key in ignore:
                continue
            if isinstance(val, (int, float, bytes, bool, str, set, list, tuple)):
                simple_vars[key] = val
        return simple_vars

    def get_pytorch_tensors(self, ignore=()):
        """
        Returns all torch.tensors in the experiment in a dict.

        Args:
            ignore (list or tuple): Iterable of names which will be ignored

        Returns:
            dict: Dictionary of PyTorch tensor

        """

        pytorch_vars = dict()
        for key, val in self.__dict__.items():
            if key in ignore:
                continue
            if torch.is_tensor(val):
                pytorch_vars[key] = val
        return pytorch_vars

    def get_pytorch_variables(self, ignore=()):
        """Same as :meth:`.get_pytorch_tensors`."""
        return self.get_pytorch_tensors(ignore)

    def save_results(self, name="results.json"):
        """
        Saves the result dict as a json file in the result dir of the experiment logger.

        Args:
            name (str): The name of the json file in which the results are written.

        """
        if self.elog is None:
            return
        with open(os.path.join(self.elog.result_dir, name), "w") as file_:
            json.dump(self.results, file_, indent=4)

    def save_pytorch_models(self):
        """Saves all torch.nn.Modules as model files in the experiment checkpoint folder."""

        if self.elog is None:
            return

        pyth_modules = self.get_pytorch_modules()
        for key, val in pyth_modules.items():
            self.elog.save_model(val, key)

    def load_pytorch_models(self):
        """Loads all model files from the experiment checkpoint folder."""

        if self.elog is None:
            return
        pyth_modules = self.get_pytorch_modules()
        for key, val in pyth_modules.items():
            self.elog.load_model(val, key)

    def log_simple_vars(self):
        """
        Logs all simple python member variables as a json file in the experiment log folder.
        The file will be names 'simple_vars.json'.
        """

        if self.elog is None:
            return
        simple_vars = self.get_simple_variables()
        with open(os.path.join(self.elog.log_dir, "simple_vars.json"), "w") as file_:
            json.dump(simple_vars, file_)

    def load_simple_vars(self):
        """
        Restores all simple python member variables from the 'simple_vars.json' file in the log
        folder.
        """

        if self.elog is None:
            return
        simple_vars = {}
        with open(os.path.join(self.elog.log_dir, "simple_vars.json"), "r") as file_:
            simple_vars = json.load(file_)
        self.update_attributes(simple_vars)

    def save_checkpoint(self,
                        name="checkpoint",
                        save_types=("model", "optimizer", "simple", "th_vars", "results"),
                        n_iter=None,
                        iter_format="{:05d}",
                        prefix=False):
        """
        Saves a current model checkpoint from the experiment.

        Args:
            name (str): The name of the checkpoint file
            save_types (list or tuple): What kind of member variables should be stored? Choices are:
                "model" <-- Pytorch models,
                "optimizer" <-- Optimizers,
                "simple" <-- Simple python variables (basic types and lists/tuples),
                "th_vars" <-- torch tensors,
                "results" <-- The result dict
            n_iter (int): Number of iterations. Together with the name, defined by the iter_format,
                a file name will be created.
            iter_format (str): Defines how the name and the n_iter will be combined.
            prefix (bool): If True, the formatted n_iter will be prepended, otherwise appended.

        """

        if self.elog is None:
            return

        model_dict = {}
        optimizer_dict = {}
        simple_dict = {}
        th_vars_dict = {}
        results_dict = {}

        if "model" in save_types:
            model_dict = self.get_pytorch_modules()
        if "optimizer" in save_types:
            optimizer_dict = self.get_pytorch_optimizers()
        if "simple" in save_types:
            simple_dict = self.get_simple_variables()
        if "th_vars" in save_types:
            th_vars_dict = self.get_pytorch_variables()
        if "results" in save_types:
            results_dict = {"results": self.results}

        checkpoint_dict = {**model_dict, **optimizer_dict, **simple_dict, **th_vars_dict, **results_dict}

        self.elog.save_checkpoint(name=name, n_iter=n_iter, iter_format=iter_format, prefix=prefix,
                                  move_to_cpu=self._checkpoint_to_cpu, **checkpoint_dict)

    def load_checkpoint(self,
                        name="checkpoint",
                        save_types=("model", "optimizer", "simple", "th_vars", "results"),
                        n_iter=None,
                        iter_format="{:05d}",
                        prefix=False,
                        path=None):
        """
        Loads a checkpoint and restores the experiment.

        Make sure you have your torch stuff already on the right devices beforehand,
        otherwise this could lead to errors e.g. when making a optimizer step
        (and for some reason the Adam states are not already on the GPU:
        https://discuss.pytorch.org/t/loading-a-saved-model-for-continue-training/17244/3 )

        Args:
            name (str): The name of the checkpoint file
            save_types (list or tuple): What kind of member variables should be loaded? Choices are:
                "model" <-- Pytorch models,
                "optimizer" <-- Optimizers,
                "simple" <-- Simple python variables (basic types and lists/tuples),
                "th_vars" <-- torch tensors,
                "results" <-- The result dict
            n_iter (int): Number of iterations. Together with the name, defined by the iter_format,
                a file name will be created and searched for.
            iter_format (str): Defines how the name and the n_iter will be combined.
            prefix (bool): If True, the formatted n_iter will be prepended, otherwise appended.
            path (str): If no path is given then it will take the current experiment dir and formatted
                name, otherwise it will simply use the path and the formatted name to define the
                checkpoint file.

        """
        if self.elog is None:
            return

        model_dict = {}
        optimizer_dict = {}
        simple_dict = {}
        th_vars_dict = {}
        results_dict = {}

        if "model" in save_types:
            model_dict = self.get_pytorch_modules()
        if "optimizer" in save_types:
            optimizer_dict = self.get_pytorch_optimizers()
        if "simple" in save_types:
            simple_dict = self.get_simple_variables()
        if "th_vars" in save_types:
            th_vars_dict = self.get_pytorch_variables()
        if "results" in save_types:
            results_dict = {"results": self.results}

        checkpoint_dict = {**model_dict, **optimizer_dict, **simple_dict, **th_vars_dict, **results_dict}

        if n_iter is not None:
            name = name_and_iter_to_filename(name,
                                             n_iter,
                                             ".pth.tar",
                                             iter_format=iter_format,
                                             prefix=prefix)

        if path is None:
            restore_dict = self.elog.load_checkpoint(name=name, **checkpoint_dict)
        else:
            checkpoint_path = os.path.join(path, name)
            if checkpoint_path.endswith("/"):
                checkpoint_path = checkpoint_path[:-1]
            restore_dict = self.elog.load_checkpoint_static(checkpoint_file=checkpoint_path, **checkpoint_dict)

        self.update_attributes(restore_dict)

    def _end_internal(self):
        """Ends the experiment and stores the final results/checkpoint"""
        if isinstance(self.results, ResultLogDict):
            self.results.close()
        self.save_results()
        self.save_end_checkpoint()
        self._save_exp_config()
        self.print("Experiment ended. Checkpoints stored =)")

    def _end_test_internal(self):
        """Ends the experiment after test and stores the final results and config"""
        self.save_results()
        self._save_exp_config()
        self.print("Testing ended. Results stored =)")

    def at_exit_func(self):
        """
        Stores the results and checkpoint at the end (if not already stored).
        This method is also called if an error occurs.
        """

        if self._exp_state not in ("Ended", "Tested"):
            if isinstance(self.results, ResultLogDict):
                self.results.print_to_file("]")
            self.save_checkpoint(name="checkpoint_exit-" + self._exp_state)
            self.save_results()
            self._save_exp_config()
            self.print("Experiment exited. Checkpoints stored =)")
        time.sleep(2)  # allow checkpoint saving to finish. We need a better solution for this :D

    def _setup_internal(self):
        self.prepare_resume()

        if self.elog is not None:
            self.elog.save_config(self._config_raw, "config")
        self._save_exp_config()

    def _start_internal(self):
        self._save_exp_config()

    def prepare_resume(self):
        """Tries to resume the experiment by using the defined resume path or PytorchExperiment."""

        checkpoint_file = ""
        base_dir = ""

        reset_epochs = self._resume_reset_epochs

        if self._resume_path is not None:
            if isinstance(self._resume_path, str):
                if self._resume_path.endswith(".pth.tar"):
                    checkpoint_file = self._resume_path
                    base_dir = os.path.dirname(os.path.dirname(checkpoint_file))
                elif self._resume_path.endswith("checkpoint") or self._resume_path.endswith("checkpoint/"):
                    checkpoint_file = get_last_file(self._resume_path)
                    base_dir = os.path.dirname(os.path.dirname(checkpoint_file))
                elif "checkpoint" in os.listdir(self._resume_path) and "config" in os.listdir(self._resume_path):
                    checkpoint_file = get_last_file(self._resume_path)
                    base_dir = self._resume_path
                else:
                    warnings.warn("You have not selected a valid experiment folder, will search all sub folders",
                                  UserWarning)
                    if self.elog is not None:
                        self.elog.text_logger.log_to("You have not selected a valid experiment folder, will search all "
                                                     "sub folders", "warnings")
                    checkpoint_file = get_last_file(self._resume_path)
                    base_dir = os.path.dirname(os.path.dirname(checkpoint_file))

        if base_dir:
            if not self._ignore_resume_config:
                load_config = Config()
                load_config.load(os.path.join(base_dir, "config/config.json"))
                self._config_raw = load_config
                self.config = Config.init_objects(self._config_raw)
                self.print("Loaded existing config from:", base_dir)
                if self.n_epochs is None:
                    self.n_epochs = self._config_raw.get("n_epochs")

        if checkpoint_file:
            self.load_checkpoint(name="", path=checkpoint_file, save_types=self._resume_save_types)
            self._resume_path = checkpoint_file
            shutil.copyfile(checkpoint_file, os.path.join(self.elog.checkpoint_dir, "0_checkpoint.pth.tar"))
            self.print("Loaded existing checkpoint from:", checkpoint_file)

            self._resume_reset_epochs = reset_epochs
            if self._resume_reset_epochs:
                self._epoch_idx = 0

    def _end_epoch_internal(self, epoch):
        self.save_results()
        if epoch % self._save_checkpoint_every_epoch == 0:
            self.save_temp_checkpoint()
        self._save_exp_config()

    def _save_exp_config(self):

        if self.elog is not None:
            cur_time = time.strftime("%y-%m-%d_%H:%M:%S", time.localtime(time.time()))
            self.elog.save_config(Config(**{'name': self.exp_name,
                                            'time': self._time_start,
                                            'state': self._exp_state,
                                            'current_time': cur_time,
                                            'epoch': self._epoch_idx
                                            }),
                                  "exp")

    def save_temp_checkpoint(self):
        """Saves the current checkpoint as checkpoint_current."""
        self.save_checkpoint(name="checkpoint_current")

    def save_end_checkpoint(self):
        """Saves the current checkpoint as checkpoint_last."""
        self.save_checkpoint(name="checkpoint_last")

    def add_result(self, value, name, counter=None, tag=None, label=None, plot_result=True, plot_running_mean=False):
        """
        Saves a results and add it to the result dict, this is similar to results[key] = val,
        but in addition also logs the value to the combined logger
        (it also stores in the results-logs file).

        **This should be your preferred method to log your numeric values**

        Args:
            value: The value of your variable
            name (str): The name/key of your variable
            counter (int or float): A counter which can be seen as the x-axis of your value.
                Normally you would just use the current epoch for this.
            tag (str): A label/tag which can group similar values and will plot values with the same
                label in the same plot
            label: deprecated label
            plot_result (bool): By default True, will also log all your values to the combined
                logger (with show_value).

        """

        if label is not None:
            warnings.warn("label in add_result is deprecated, please use tag instead")

            if tag is None:
                tag = label

        tag_name = tag
        if tag_name is None:
            tag_name = name

        r_elem = ResultElement(data=value, label=tag_name, epoch=self._epoch_idx, counter=counter)

        self.results[name] = r_elem

        if plot_result:
            if tag is None:
                legend = False
            else:
                legend = True
            if plot_running_mean:
                value = np.mean(self.results.running_mean_dict[name])
            self.clog.show_value(value=value, name=name, tag=tag_name, counter=counter, show_legend=legend)

    def get_result(self, name):
        """
        Similar to result[key] this will return the values in the results dictionary with the given
        name/key.

        Args:
            name (str): the name/key for which a value is stored.

        Returns:
            The value with the key 'name' in the results dict.

        """
        return self.results.get(name)

    def add_result_without_epoch(self, val, name):
        """
        A faster method to store your results, has less overhead and does not call the combined
        logger. Will only store to the results dictionary.

        Args:
            val: the value you want to add.
            name (str): the name/key of your value.

        """
        self.results[name] = val

    def get_result_without_epoch(self, name):
        """
        Similar to result[key] this will return the values in result with the given name/key.

        Args:
            name (str): the name/ key for which a value is stores.

        Returns:
            The value with the key 'name' in the results dict.

        """
        return self.results.get(name)

    def print(self, *args):
        """
        Calls 'print' on the experiment logger or uses builtin 'print' if former is not
        available.
        """

        if self.elog is None:
            print(*args)
        else:
            self.elog.print(*args)


def get_last_file(dir_, name=None):
    """
    Returns the (alphabetically) last file in the folder which matches the name supplied

    Args:
        dir_: The base directory to start the search in
        name: The name pattern to match with the files

    Returns:
        str: the path to the (alphabetically) last file

    """
    if name is None:
        name = "*checkpoint*.pth.tar"

    dir_files = []

    for root, dirs, files in os.walk(dir_):
        for filename in fnmatch.filter(files, name):
            if 'last' in filename:
                return os.path.join(root, filename)
            checkpoint_file = os.path.join(root, filename)
            dir_files.append(checkpoint_file)

    if len(dir_files) == 0:
        return ""

    last_file = sorted(dir_files, reverse=True)[0]

    return last_file


def get_vars_from_sys_argv():
    """
    Parses the command line args (argv) and looks for --config_path and --resume_path and returns them if found.

    Returns:
        tuple: a tuple of (config_path, resume_path ) , None if it is not found

    """
    import sys
    import argparse

    if len(sys.argv) > 1:

        parser = argparse.ArgumentParser()

        # parse just config keys
        parser.add_argument("config_path", type=str)
        parser.add_argument("resume_path", type=str)

        # parse args
        param, unknown = parser.parse_known_args()

        if len(unknown) > 0:
            warnings.warn("Called with unknown arguments: %s" % unknown, RuntimeWarning)

        # update dict
        return param.get("config_path"), param.get("resume_path")


def experimentify(setup_fn="setup", train_fn="train", validate_fn="validate", end_fn="end", test_fn="test", **decoargs):
    """
    Experimental decorator which monkey patches your class into a PytorchExperiment.
    You can then call run on your new :class:`.PytorchExperiment` class.

    Args:
        setup_fn: The name of your setup() function
        train_fn: The name of your train() function
        validate_fn: The name of your validate() function
        end_fn: The name of your end() function
        test_fn: The name of your test() function

    """

    def wrap(cls):

        ### Initilaize both Classes (as original class)
        prev_init = cls.__init__

        def new_init(*args, **kwargs):
            prev_init(*args, **kwargs)
            kwargs.update(decoargs)
            PytorchExperiment.__init__(*args, **kwargs)

        cls.__init__ = new_init

        ### Set new Experiment methods
        if not hasattr(cls, "setup") and hasattr(cls, setup_fn):
            setattr(cls, "setup", getattr(cls, setup_fn))
        elif hasattr(cls, "setup") and setup_fn != "setup":
            warnings.warn("Found already exisiting setup function in class, so will use the exisiting one")

        if not hasattr(cls, "train") and hasattr(cls, train_fn):
            setattr(cls, "train", getattr(cls, train_fn))
        elif hasattr(cls, "train") and setup_fn != "train":
            warnings.warn("Found already exisiting train function in class, so will use the exisiting one")

        if not hasattr(cls, "validate") and hasattr(cls, validate_fn):
            setattr(cls, "validate", getattr(cls, validate_fn))
        elif hasattr(cls, "validate") and setup_fn != "validate":
            warnings.warn("Found already exisiting validate function in class, so will use the exisiting one")

        if not hasattr(cls, "end") and hasattr(cls, end_fn):
            setattr(cls, "end", getattr(cls, end_fn))
        elif hasattr(cls, "end") and end_fn != "end":
            warnings.warn("Found already exisiting end function in class, so will use the exisiting one")

        if not hasattr(cls, "test") and hasattr(cls, test_fn):
            setattr(cls, "test", getattr(cls, test_fn))
        elif hasattr(cls, "test") and test_fn != "test":
            warnings.warn("Found already exisiting test function in class, so will use the exisiting one")

        ### Copy methods from PytorchExperiment into the original class
        for elem in dir(PytorchExperiment):
            if not hasattr(cls, elem):
                trans_fn = getattr(PytorchExperiment, elem)
                setattr(cls, elem, trans_fn)

        return cls

    return wrap



import torch
from trixi.logger.tensorboard.tensorboardxlogger import TensorboardXLogger


class PytorchTensorboardXLogger(TensorboardXLogger):
    """Abstract interface for visual logger."""

    def process_params(self, f, *args, **kwargs):
        """
        Inherited "decorator": convert PyTorch variables and Tensors to numpy arrays
        """

        # convert args
        args = (a.detach().cpu().numpy() if torch.is_tensor(a) else a for a in args)

        # convert kwargs
        for key, data in kwargs.items():
            if torch.is_tensor(data):
                kwargs[key] = data.detach().cpu().numpy()

        return f(self, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        super(PytorchTensorboardXLogger, self).__init__(*args, **kwargs)

    def plot_model_structure(self, model, input_size):
        """
         Plots the model structure/ model graph of a pytorch module.

         Args:
             model: The graph of this model will be plotted.
             input_size: Input size of the model (with batch dim).
         """
        inpt_vars = [torch.randn(i_s) for i_s in input_size]

        self.writer.add_graph(model=model, input_to_model=inpt_vars)



import os.path
import re
import subprocess
import subprocess as subp
import sys
import zipfile


class SourcePacker(object):
    """
    Inspired by https://github.com/IDSIA/sacred
    """

    @staticmethod
    def join_paths(*parts):
        """Join different parts together to a valid dotted path."""
        return '.'.join(str(p).strip('.') for p in parts if p)

    @staticmethod
    def iter_prefixes(path):
        """
        Iterate through all (non-empty) prefixes of a dotted path.
        Example:
        >>> list(iter_prefixes('foo.bar.baz'))
        ['foo', 'foo.bar', 'foo.bar.baz']
        """
        split_path = path.split('.')
        for i in range(1, len(split_path) + 1):
            yield SourcePacker.join_paths(*split_path[:i])

    @staticmethod
    def create_source_or_dep(mod, sources):
        filename = ''
        if mod is not None and hasattr(mod, '__file__'):
            filename = os.path.abspath(mod.__file__)

        ### To source or dependency
        if filename and filename not in sources and SourcePacker.is_source(filename):
            sources.add(filename)

    @staticmethod
    def is_source(filename):
        if ".virtualenvs" in filename or "site-packages" in filename or re.search("python[0-9]\.[0-9]", filename) is not \
                None:
            return False
        else:
            return True

    @staticmethod
    def git_info(file_):

        old_dir = os.getcwd()
        file_path = os.path.abspath(file_)
        os.chdir(os.path.dirname(file_path))

        try:
            commit = subp.check_output(["git", "rev-parse", "HEAD"]).decode("ascii")[:-1]
            branch = subp.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).decode("ascii")[:-1]
            repo = subp.check_output(["git", "remote", "-vv"]).decode("ascii")
            repo = re.findall("(?<=origin[\s\t])(http.+|ssh.+|git.+)(?=[\s\t]\(fetch)", repo)[0]
            result = (repo, branch, commit)
        except Exception as e:
            print("Could not find git info for {}".format(file_path))
            print(e)
            result = (None, None, None)

        os.chdir(old_dir)
        return result

    @staticmethod
    def gather_sources_and_dependencies(globs):
        py_str = "python {}".format(sys.version)
        dependencies = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze']).decode("utf-8").split("\n")

        filename = globs.get('__file__')

        if filename is None:
            sources = set()
        else:
            sources = set()
            sources.add(filename)
        for glob in globs.values():
            if isinstance(glob, type(sys)):
                mod_path = glob.__name__
            elif hasattr(glob, '__module__'):
                mod_path = glob.__module__
            else:
                continue

            if not mod_path:
                continue

            for modname in SourcePacker.iter_prefixes(mod_path):
                mod = sys.modules.get(modname)
                SourcePacker.create_source_or_dep(mod, sources)

        return py_str, sources, dependencies

    @staticmethod
    def zip_sources(globs, filename):

        py_str, sources, dependencies = SourcePacker.gather_sources_and_dependencies(globs=globs)
        repo, branch, commit = SourcePacker.git_info(globs.get("__file__"))
        cmd = " ".join(sys.argv)

        with zipfile.ZipFile(filename, mode='w') as zf:
            for source in sources:
                zf.write(source)

            zf.writestr("python_version.txt", py_str)
            dep_str = "\n".join(dependencies)
            zf.writestr("modules.txt", dep_str)
            git_str = "Repository: {}\nBranch: {}\nCommit: {}".format(repo, branch, commit)
            zf.writestr("git_info.txt", git_str)
            zf.writestr("command.txt", cmd)



from trixi.logger.tensorboard.tensorboardxlogger import TensorboardXLogger
try:
    from trixi.logger.tensorboard.pytorchtensorboardxlogger import PytorchTensorboardXLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))



import atexit
import tempfile
import warnings
from multiprocessing import Process

import matplotlib.pyplot as plt
import numpy as np
import torch

from torch.autograd import Variable
from torchvision.utils import make_grid

from trixi.util.util import np_make_grid, get_tensor_embedding
from trixi.logger.experiment.pytorchexperimentlogger import PytorchExperimentLogger
from trixi.logger.visdom.numpyvisdomlogger import NumpyVisdomLogger
from trixi.logger.abstractlogger import convert_params
from trixi.util.pytorchutils import get_guided_image_gradient, get_smooth_image_gradient, get_vanilla_image_gradient


class PytorchVisdomLogger(NumpyVisdomLogger):
    """
    Visual logger, inherits the NumpyVisdomLogger and plots/ logs pytorch tensors and variables on a Visdom server.
    """

    def __init__(self, *args, **kwargs):
        super(PytorchVisdomLogger, self).__init__(*args, **kwargs)

    def process_params(self, f, *args, **kwargs):
        """
        Inherited "decorator": convert Pytorch variables and Tensors to numpy arrays.
        """

        ### convert args
        args = (a.detach().cpu().numpy() if torch.is_tensor(a) else a for a in args)
        # args = (a.data.cpu().numpy() if isinstance(a, Variable) else a for a in args)

        ### convert kwargs
        for key, data in kwargs.items():
            # if isinstance(data, Variable):
            #     kwargs[key] = data.detach().cpu().numpy()
            if torch.is_tensor(data):
                kwargs[key] = data.detach().cpu().numpy()

        return f(self, *args, **kwargs)

    def plot_model_statistics(self, model, env_appendix="", model_name="", plot_grad=False, **kwargs):
        """
        Plots statstics (mean, std, abs(max)) of the weights or the corresponding gradients of a model as a barplot.

        Args:
            model: Model with the weights.
            env_appendix: Visdom environment name appendix
            model_name: Name of the model (is used as window name).
            plot_grad: If false plots weight statistics, if true plot the gradients of the weights.
        """

        means = []
        stds = []
        maxmin = []
        legendary = []
        for i, (m_param_name, m_param) in enumerate(model.named_parameters()):
            win_name = "%s_params" % model_name
            if plot_grad:
                m_param = m_param.grad
                win_name = "%s_grad" % model_name

            if m_param is not None:
                param_mean = m_param.detach().mean()
                param_std = m_param.detach().std()

                if np.isnan(param_std):
                    param_std = 0

                means.append(param_mean)
                stds.append(param_std)
                maxmin.append(torch.max(torch.abs(m_param)).item())
                legendary.append("%s-%s" % (model_name, m_param_name))

        self.show_barplot(name=win_name, array=np.asarray([means, stds, maxmin]), legend=legendary,
                          rownames=["mean", "std", "max"], env_appendix=env_appendix)

    def plot_model_statistics_weights(self, model, env_appendix="", model_name="", **kwargs):
        """
        Plots statstics (mean, std, abs(max)) of the weights of a model as a barplot (uses plot model statistics with plot_grad=False).

        Args:
            model: Model with the weights.
            env_appendix: Visdom environment name appendix
            model_name: Name of the model (is used as window name).
        """
        self.plot_model_statistics(model=model, env_appendix=env_appendix, model_name=model_name, plot_grad=False)

    def plot_model_statistics_grads(self, model, env_appendix="", model_name="", **kwargs):
        """
        Plots statstics (mean, std, abs(max)) of the gradients of a model as a barplot (uses plot model statistics with plot_grad=True).

        Args:
            model: Model with the weights and the corresponding gradients (have to calculated previously).
            env_appendix: Visdom environment name appendix
            model_name: Name of the model (is used as window name).
        """
        self.plot_model_statistics(model=model, env_appendix=env_appendix, model_name=model_name, plot_grad=True)

    def plot_model_gradient_flow(self, model, name="model", title=None):
        """
        Plots statstics (mean, std, abs(max)) of the weights or the corresponding gradients of a model as a barplot.

        Args:
            model: Model with the weights.
            env_appendix: Visdom environment name appendix, if none is given, it uses "-histogram".
            model_name: Name of the model (is used as window name).
            plot_grad: If false plots weight statistics, if true plot the gradients of the weights.
        """
        ave_grads = []
        layers = []

        named_parameters = model.named_parameters()
        for n, p in named_parameters:
            if (p.requires_grad) and ("bias" not in n):
                layers.append(n)
                ave_grads.append(p.grad.abs().mean())

        plt.figure()
        plt.plot(ave_grads, alpha=0.3, color="b")
        plt.hlines(0, 0, len(ave_grads) + 1, linewidth=1, color="k")
        plt.xticks(range(0, len(ave_grads), 1), layers, rotation="vertical")
        plt.xlim(xmin=0, xmax=len(ave_grads))
        plt.xlabel("Layers")
        plt.ylabel("average gradient {}".format(name))
        plt.title("Gradient flow")
        plt.grid(True)

        self.show_matplot_plt(plt.gcf(), name=name, title=title)

    def plot_mutliple_models_statistics_weights(self, model_dict, env_appendix=None, **kwargs):
        """
        Given models in a dict, plots the weight statistics of the models.

        Args:
            model_dict: Dict with models, the key is assumed to be the name, while the value is the model.
            env_appendix: Visdom environment name appendix
        """
        for model_name, model in model_dict.items():
            self.plot_model_statistics_weights(model=model, env_appendix=env_appendix, model_name=model_name)

    def plot_mutliple_models_statistics_grads(self, model_dict, env_appendix="", **kwargs):
        """
        Given models in a dict, plots the gradient statistics of the models.

        Args:
            model_dict: Dict with models, the key is assumed to be the name, while the value is the model.
            env_appendix: Visdom environment name appendix
        """
        for model_name, model in model_dict.items():
            self.plot_model_statistics_grads(model=model, env_appendix=env_appendix, model_name=model_name)

    def plot_model_structure(self, model, input_size, name=None, use_cuda=True, delete_tmp_on_close=False, **kwargs):
        """
        Plots the model structure/ model graph of a pytorch module (this only works correctly with pytorch 0.2.0).

        Args:
            model: The graph of this model will be plotted.
            input_size: Input size of the model (with batch dim).
            name: The name of the window in the visdom env.
            use_cuda: Perform model dimension calculations on the gpu (cuda).
            delete_tmp_on_close: Determines if the tmp file will be deleted on close. If set true, can cause problems due to the multi threadded plotting.
        """

        if not torch.cuda.is_available():
            use_cuda = False

        def make_dot(output_var, state_dict=None):
            """
            Produces Graphviz representation of Pytorch autograd graph.
            Blue nodes are the Variables that require grad, orange are Tensors
            saved for backward in torch.autograd.Function.

            Args:
                output_var: output Variable
                state_dict: dict of (name, parameter) to add names to node that require grad
            """
            from graphviz import Digraph

            if state_dict is not None:
                # assert isinstance(params.values()[0], Variable)
                param_map = {id(v): k for k, v in state_dict.items()}

            node_attr = dict(style='filled',
                             shape='box',
                             align='left',
                             fontsize='12',
                             ranksep='0.1',
                             height='0.2')
            dot = Digraph(node_attr=node_attr, graph_attr=dict(size="12,12"), format="svg")
            seen = set()

            def size_to_str(size):
                return '(' + (', ').join(['%d' % v for v in size]) + ')'

            def add_nodes(var):
                if var not in seen:
                    if torch.is_tensor(var):
                        dot.node(str(id(var)), size_to_str(var.size()), fillcolor='orange')
                    elif hasattr(var, 'variable'):
                        u = var.variable
                        if state_dict is not None and id(u.data) in param_map:
                            node_name = param_map[id(u.data)]
                        else:
                            node_name = ""
                        node_name = '%s\n %s' % (node_name, size_to_str(u.size()))
                        dot.node(str(id(var)), node_name, fillcolor='lightblue')
                    else:
                        node_name = str(type(var).__name__)
                        if node_name.endswith("Backward"):
                            node_name = node_name[:-8]
                        dot.node(str(id(var)), node_name)
                    seen.add(var)
                    if hasattr(var, 'next_functions'):
                        for u in var.next_functions:
                            if u[0] is not None:
                                dot.edge(str(id(u[0])), str(id(var)))
                                add_nodes(u[0])
                    if hasattr(var, 'saved_tensors'):
                        for t in var.saved_tensors:
                            dot.edge(str(id(t)), str(id(var)))
                            add_nodes(t)

            add_nodes(output_var.grad_fn)
            return dot

        # Create input
        inpt_vars = [torch.randn(i_s) for i_s in input_size]
        if use_cuda:
            inpt_vars = [i_v.cuda() for i_v in inpt_vars]
            model = model.cuda()

        # get output
        output = model(*inpt_vars)

        # get temp file to store svg in
        fp = tempfile.NamedTemporaryFile(suffix=".svg", delete=delete_tmp_on_close)
        g = make_dot(output, model.state_dict())

        try:

            # Create model graph and store it as svg
            x = g.render(fp.name[:-4], cleanup=True)

            # Display model graph in visdom
            self.show_svg(svg=x, name=name)
        except Exception as e:
            warnings.warn("Could not render model, make sure the Graphviz executables are on your system.")

    def show_image_grid(self, images, name=None, caption=None, env_appendix="", opts=None,
                        image_args=None, **kwargs):
        """
        Calls the save image grid method (for abstract logger combatibility)

        Args:
           images: 4d- tensor (N, C, H, W)
           name: The name of the window
           caption: Caption of the generated image grid
           env_appendix: appendix to the environment name, if used the new env is env+env_appendix
           opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
           image_args: Arguments for the tensorvision save image method


        """

        if opts is None: opts = {}
        if image_args is None: image_args = {}

        tensor = images.detach().cpu()
        viz_task = {
            "type": "image_grid",
            "tensor": tensor,
            "name": name,
            "caption": caption,
            "env_appendix": env_appendix,
            "opts": opts,
            "image_args": image_args
        }
        self._queue.put_nowait(viz_task)

    def __show_image_grid(self, tensor, name=None, caption=None, env_appendix="", opts=None,
                          image_args=None, **kwargs):
        """
          Internal show_image_grid method, called by the internal process.
          This function does all the magic.
        """

        if opts is None: opts = {}
        if image_args is None: image_args = {}

        if isinstance(tensor, Variable):

            tensor = tensor.detach()

        if torch.is_tensor(tensor):
            assert torch.is_tensor(tensor), "tensor has to be a pytorch tensor or variable"
            assert tensor.dim() == 4, "tensor has to have 4 dimensions"
            if not (tensor.size(1) == 1 or tensor.size(1) == 3):
                warnings.warn("The 1. dimension (channel) has to be either 1 (gray) or 3 (rgb), taking the first "
                              "dimension now !!!")
                tensor = tensor[:, 0:1, ]

            grid = make_grid(tensor, **image_args)
            image = grid.mul(255).clamp(0, 255).byte().numpy()
        elif isinstance(tensor, np.ndarray):
            grid = np_make_grid(tensor, **image_args)
            image = np.clip(grid * 255, a_min=0, a_max=255)
            image = image.astype(np.uint8)

        else:
            raise ValueError("Tensor has to be a torch tensor or a numpy array")

        opts = opts.copy()
        opts.update(dict(
            title=name,
            caption=caption
        ))

        win = self.vis.image(
            img=image,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    NumpyVisdomLogger.show_funcs["image_grid"] = __show_image_grid

    @convert_params
    def show_image_grid_heatmap(self, heatmap, background=None, ratio=0.3, normalize=True, colormap=2,
                                name=None, caption=None, env_appendix="", opts=None,
                                image_args=None, **kwargs):
        """
        Creates heat map from the given map and if given combines it with the background and then
        displays results with as image grid.

        Args:
           heatmap:  4d- tensor (N, C, H, W)
           background: 4d- tensor (N, C, H, W)
           name: The name of the window
           ratio: The ratio to mix the map with the background (0 = only background, 1 = only map)
           caption: Caption of the generated image grid
           env_appendix: appendix to the environment name, if used the new env is env+env_appendix
           opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
           image_args: Arguments for the tensorvision save image method

        """

        if opts is None: opts = {}
        if image_args is None: image_args = {}

        viz_task = {
            "type": "image_grid_heatmap",
            "tensor": background,
            "heatmap": heatmap,
            "ratio": ratio,
            "normalize": normalize,
            "colormap": colormap,
            "name": name,
            "caption": caption,
            "env_appendix": env_appendix,
            "opts": opts,
            "image_args": image_args
        }
        self._queue.put_nowait(viz_task)

    def __show_image_grid_heatmap(self, heatmap, tensor=None, ratio=0.3, colormap=2,
                                  normalize=True, name=None, caption=None,
                                  env_appendix="", opts=None, image_args=None, **kwargs):
        """
          Internal show_image_grid_heatmap method, called by the internal process.
          This function does all the magic.
        """

        from cv2 import cv2

        if opts is None: opts = {}
        if image_args is None: image_args = {}

        map_grid = np_make_grid(heatmap, normalize=normalize)
        map_ = np.clip(map_grid * 255, a_min=0, a_max=255)
        map_ = map_.astype(np.uint8)

        map_ = cv2.applyColorMap(map_.transpose(1, 2, 0), colormap=colormap)
        map_ = cv2.cvtColor(map_, cv2.COLOR_BGR2RGB)
        map_ = map_.transpose(2, 0, 1)

        fuse_img = map_

        if tensor is not None:
            img_grid = np_make_grid(tensor, **image_args)
            image = np.clip(img_grid * 255, a_min=0, a_max=255)
            image = image.astype(np.uint8)

            fuse_img = (1.0 - ratio) * image + ratio * map_

        opts = opts.copy()
        opts.update(dict(
            title=name,
            caption=caption
        ))

        win = self.vis.image(
            img=fuse_img,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    NumpyVisdomLogger.show_funcs["image_grid_heatmap"] = __show_image_grid_heatmap

    @convert_params
    def show_embedding(self, tensor, labels=None, name=None, method="tsne", n_dims=2, n_neigh=30, meth_args=None,
                       *args, **kwargs):
        """
        Displays a tensor a an embedding

        Args:
            tensor: Tensor to be embedded an then displayed
            labels: Labels of the entries in the tensor (first dimension)
            name: The name of the window
            method: Method used for embedding, options are: tsne, standard, ltsa, hessian, modified, isomap, mds,
            spectral, umap
            n_dims: dimensions to embed the data into
            n_neigh: Neighbour parameter to kind of determin the embedding (see t-SNE for more information)
            meth_args: Further arguments which can be passed to the embedding method

        """

        if meth_args is None:
            meth_args = {}

        def __show_embedding(queue, tensor, labels=None, name=None, method="tsne", n_dims=2, n_neigh=30, **meth_args):
            emb_data = get_tensor_embedding(tensor, method=method, n_dims=n_dims, n_neigh=n_neigh, **meth_args)

            vis_task = {
                "type": "scatterplot",
                "array": emb_data,
                "labels": labels,
                "name": name,
                "env_appendix": "",
                "opts": {}
            }
            queue.put_nowait(vis_task)

        p = Process(target=__show_embedding, kwargs=dict(queue=self._queue,
                                                         tensor=tensor,
                                                         labels=labels,
                                                         name=name,
                                                         method=method,
                                                         n_dims=n_dims,
                                                         n_neigh=n_neigh,
                                                         **meth_args
                                                         ))
        atexit.register(p.terminate)
        p.start()

    @convert_params
    def show_roc_curve(self, tensor, labels, name, reduce_to_n_samples=None, use_sub_process=False):
        """
        Displays a roc curve given a tensor with scores and the coresponding labels

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            name: The name of the window
            reduce_to_n_samples: Reduce/ downsample to to n samples for fewer data points
            use_sub_process: Use a sub process to do the processing

        """

        res_fn = lambda tpr, fpr: self.show_lineplot(tpr, fpr, name=name, opts={"fillarea": True,
                                                                                "webgl": True})
        PytorchExperimentLogger.get_roc_curve(tensor=tensor, labels=labels, reduce_to_n_samples=reduce_to_n_samples,
                                              use_sub_process=use_sub_process, results_fn=res_fn)

    @convert_params
    def show_pr_curve(self, tensor, labels, name, reduce_to_n_samples=None, use_sub_process=False):
        """
        Displays a precision recall curve given a tensor with scores and the coresponding labels

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            name: The name of the window
            reduce_to_n_samples: Reduce/ downsample to to n samples for fewer data points
            use_sub_process: Use a sub process to do the processing
        """
        res_fn = lambda precision, recall: self.show_lineplot(precision, recall, name=name, opts={"fillarea": True,
                                                                                                  "webgl": True})
        PytorchExperimentLogger.get_pr_curve(tensor=tensor, labels=labels, reduce_to_n_samples=reduce_to_n_samples,
                                             use_sub_process=use_sub_process, results_fn=res_fn)

    @convert_params
    def show_classification_metrics(self, tensor, labels, name, metric=("roc-auc", "pr-score"),
                                    use_sub_process=False, tag_name=None):
        """
        Displays some classification metrics as line plots in a graph (similar to show value (also uses show value
        for the caluclated values))

        Args:
            tensor: Tensor with scores (e.g class probability )
            labels: Labels of the samples to which the scores match
            name: The name of the window
            metric: List of metrics to calculate. Options are: roc-auc, pr-auc, pr-score, mcc, f1

        Returns:

        """

        res_fn = lambda val, name, tag: self.show_value(val, name=name, tag=tag)
        PytorchExperimentLogger.get_classification_metrics(tensor=tensor, labels=labels, name=name, metric=metric,
                                                           use_sub_process=use_sub_process, tag_name=tag_name,
                                                           results_fn=res_fn)

    def show_image_gradient(self, model, inpt, err_fn, grad_type="vanilla", n_runs=20, eps=0.1,
                            abs=False, **image_grid_params):
        """
        Given a model creates calculates the error and backpropagates it to the image and saves it (saliency map).

        Args:
            model: The model to be evaluated
            inpt: Input to the model
            err_fn: The error function the evaluate the output of the model on
            grad_type: Gradient calculation method, currently supports (vanilla, vanilla-smooth, guided,
            guided-smooth) ( the guided backprob can lead to segfaults -.-)
            n_runs: Number of runs for the smooth variants
            eps: noise scaling to be applied on the input image (noise is drawn from N(0,1))
            abs (bool): Flag, if the gradient should be a absolute value
            **image_grid_params: Params for make image grid.

        """
        grad = PytorchExperimentLogger.get_input_gradient(model=model, inpt=inpt, err_fn=err_fn, grad_type=grad_type,
                                                          n_runs=n_runs, eps=eps, abs=abs)
        self.show_image_grid(grad, **image_grid_params)



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import seaborn

from trixi.logger.abstractlogger import AbstractLogger, convert_params
from trixi.util.util import chw_to_hwc


class NumpySeabornPlotLogger(AbstractLogger):
    """
    Visual logger, inherits the AbstractLogger and plots/ logs numpy arrays/ values as matplotlib / seaborn plots.
    """

    def __init__(self, **kwargs):
        """
        Initializes a new NumpySeabornPlotLogger
        """
        super(NumpySeabornPlotLogger, self).__init__(**kwargs)

        self.values = defaultdict(lambda: defaultdict(list))
        self.max_values = defaultdict(int)

    @convert_params
    def show_image(self, image, name=None, show=True, *args, **kwargs):
        """
        Create an image figure

        Args:
            image: The image array to be displayed
            name: The name of the image window
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure

        """
        figure = self.get_figure(name)
        plt.clf()

        image = chw_to_hwc(image)

        plt.imshow(image)
        plt.axis("off")
        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    @convert_params
    def show_value(self, value, name, counter=None, tag=None, show=True, *args, **kwargs):
        """
       Creates a line plot that is automatically appended with new values and returns it as a figure.

       Args:
           value: Value to be plotted / appended to the graph (y-axis value)
           name: The name of the window
           counter: counter, which tells the number of the sample (with the same name) (x-axis value)
           tag: Tag, grouping similar values. Values with the same tag will be plotted in the same plot
           show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure

               """
        figure = self.get_figure(tag)
        plt.clf()

        seaborn.set_style("whitegrid")

        if tag is None:
            tag = name

        if counter is None:
            counter = len(self.values[tag][name]) + 1

        max_val = max(self.max_values[tag], counter)
        self.max_values[tag] = max_val
        self.values[tag][name].append((value, max_val))

        for y_tag in self.values[tag]:

            y, x = zip(*self.values[tag][y_tag])
            plt.plot(x, y, label=y_tag)

        if show:
            plt.show(block=False)
            plt.pause(0.01)

        plt.legend()
        plt.ylabel(name)

        return figure

    @convert_params
    def show_barplot(self, array, name=None, show=True, *args, **kwargs):
        """
        Creates a bar plot figure from an array

        Args:
            array: array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name: The name of the figure
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure
        """

        figure = self.get_figure(name)

        y = array
        x = list(range(1, len(y) + 1))

        seaborn.set_style("whitegrid")

        ax = seaborn.barplot(y=y, x=x)

        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    @convert_params
    def show_boxplot(self, array, name, show=True, *args, **kwargs):
        """
        Creates a box plot figure from an array

        Args:
            array: array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name: The name of the figure
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure
        """

        figure = self.get_figure(name)
        seaborn.set_style("whitegrid")

        ax = seaborn.boxplot(data=array)

        handles, _ = ax.get_legend_handles_labels()
        try:
            legend = kwargs['opts']['legend']
            ax.legend(handles, legend)
        except KeyError: # if no legend is defined
            pass
        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    @convert_params
    def show_lineplot(self, y_vals, x_vals=None, name=None, show=True, *args, **kwargs):
        """
        Creates a line plot figure with (multiple) lines plot, given values Y (and optional the corresponding X values)

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
            not set the points are assumed to be equally distributed in the interval [0, 1] )
            name: The name of the figure
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure
        """

        figure = self.get_figure(name)
        plt.clf()

        seaborn.set_style("whitegrid")

        if x_vals is None:
            x_vals = list(range(len(y_vals)))

        plt.plot(x_vals, y_vals)

        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    @convert_params
    def show_scatterplot(self, array, name=None, show=True, *args, **kwargs):
        """
        Creates a scatter plot figure with the points given in array

        Args:
            array: A 2d array with size N x dim, where each element i \in N at X[i] results in a a 2d (if dim = 2)/
            3d (if dim = 3) point.
            name: The name of the figure
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure
        """

        if not isinstance(array, np.ndarray):
            raise TypeError("Array must be numpy arrays (this class is called NUMPY seaborn logger, and seaborn"
                            " can only handle numpy arrays -.- .__. )")
        if len(array.shape) != 2:
            raise ValueError("Array must be 2D for scatterplot")
        if array.shape[1] != 2:
            raise ValueError("Array must be 2D and have x,y pairs in the 2nd dim for scatterplot")

        x, y = zip(*array)
        x, y = np.asarray(x), np.asarray(y)

        figure = self.get_figure(name)
        plt.clf()

        ax = seaborn.regplot(x=x, y=y, fit_reg=False)

        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    @convert_params
    def show_piechart(self, array, name=None, show=True, *args, **kwargs):
        """
        Creates a scatter plot figure

        Args:
            array: Array of positive integers. Each integer will be presented as a part of the pie (with the total
            as the sum of all integers)
            name: The name of the figure
            show: Flag if it should also display the figure (result might also depend on the matplotlib backend )

        Returns:
            A matplotlib figure
        """

        figure = self.get_figure(name)
        plt.clf()

        plt.pie(array)

        if show:
            plt.show(block=False)
            plt.pause(0.01)

        return figure

    def get_figure(self, name):
        """
        Returns a figure with a given name as identifier.

        If no figure yet exists with the name a new one is created. Otherwise the existing one is returned

        Args:
            name: Name of the figure

        Returns:
            A figure with the given name
        """

        return plt.figure(name)



import os
import warnings

import torch
from PIL import Image
import numpy as np
from cv2 import cv2
from scipy.misc import imsave
from torch.autograd import Variable
from torchvision.utils import save_image as tv_save_image
from trixi.util.util import np_make_grid

from trixi.logger.abstractlogger import threaded, convert_params
from trixi.logger.file.numpyplotfilelogger import NumpyPlotFileLogger
from trixi.util import name_and_iter_to_filename
from trixi.util.pytorchutils import get_guided_image_gradient, get_smooth_image_gradient, get_vanilla_image_gradient


class PytorchPlotFileLogger(NumpyPlotFileLogger):
    """
    Visual logger, inherits the NumpyPlotLogger and plots/ logs pytorch tensors and variables as files on the local
    file system.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes a PytorchPlotFileLogger to plot images, plots, ... into an image and plot directory

        Args:
            img_dir: The directory to store images in
            plot_dir: The directory to store plots in
        """
        super(PytorchPlotFileLogger, self).__init__(*args, **kwargs)

    def process_params(self, f, *args, **kwargs):
        """
        Inherited "decorator": convert Pytorch variables and Tensors to numpy arrays
        """

        ### convert args
        args = (a.detach().cpu().numpy() if torch.is_tensor(a) else a for a in args)

        ### convert kwargs
        for key, data in kwargs.items():
            if torch.is_tensor(data):
                kwargs[key] = data.detach().cpu().numpy()

        return f(self, *args, **kwargs)

    @staticmethod
    @threaded
    def save_image_static(image_dir, tensor, name, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None):
        """
        Saves an image tensor in an image directory

        Args:
            image_dir: Directory to save the image in
            tensor: Tensor containing the image
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method
        """

        if isinstance(tensor, np.ndarray):
            tensor = torch.from_numpy(tensor)

        if image_args is None:
            image_args = {}

        if n_iter is not None:
            name = name_and_iter_to_filename(name=name, n_iter=n_iter, ending=".png", iter_format=iter_format,
                                             prefix=prefix)
        elif not name.endswith(".png"):
            name = name + ".png"

        img_file = os.path.join(image_dir, name)
        os.makedirs(os.path.dirname(img_file), exist_ok=True)
        tv_save_image(tensor=tensor, filename=img_file, **image_args)

    def save_image(self, tensor, name, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None):
        """
        Saves an image into the image directory of the PytorchPlotFileLogger

        Args:
            tensor: Tensor containing the image
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method

        """

        if image_args is None:
            image_args = {}

        self.save_image_static(image_dir=self.img_dir, tensor=tensor, name=name, n_iter=n_iter, iter_format=iter_format,
                               prefix=prefix, image_args=image_args)

    @staticmethod
    @threaded
    def save_images_static(image_dir, tensors, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None):
        """
        Saves an image tensors in an image directory

        Args:
            image_dir: Directory to save the image in
            tensors: A dict with file-name-> tensor to plot as image
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method
        """

        assert isinstance(tensors, dict)
        if image_args is None:
            image_args = {}

        for name, tensor in tensors.items():
            PytorchPlotFileLogger.save_image_static(image_dir=image_dir, tensor=tensor, name=name, n_iter=n_iter,
                                                    iter_format=iter_format, prefix=prefix, image_args=image_args)

    def save_images(self, tensors, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None):
        """
        Saves an image tensors into the image directory of the PytorchPlotFileLogger

        Args:
            tensors: A dict with file-name-> tensor to plot as image
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method
        """

        assert isinstance(tensors, dict)
        if image_args is None:
            image_args = {}

        self.save_images_static(image_dir=self.img_dir, tensors=tensors, n_iter=n_iter, iter_format=iter_format,
                                prefix=prefix, image_args=image_args)

    @staticmethod
    @threaded
    def save_image_grid_static(image_dir, tensor, name, n_iter=None, prefix=False, iter_format="{:05d}",
                               image_args=None):
        """
        Saves images of a 4d- tensor (N, C, H, W) as a image grid into an image file in a given directory

        Args:
            image_dir: Directory to save the image in
            tensor: 4d- tensor (N, C, H, W)
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method

        """

        if isinstance(tensor, np.ndarray):
            tensor = torch.tensor(tensor)

        if not (tensor.size(1) == 1 or tensor.size(1) == 3):
            warnings.warn("The 1. dimension (channel) has to be either 1 (gray) or 3 (rgb), taking the first "
                          "dimension now !!!")
            tensor = tensor[:, 0:1, ]

        if n_iter is not None:
            name = name_and_iter_to_filename(name=name, n_iter=n_iter, ending=".png", iter_format=iter_format,
                                             prefix=prefix)
        elif not name.endswith(".png"):
            name += ".png"

        img_file = os.path.join(image_dir, name)

        if image_args is None:
            image_args = {}

        os.makedirs(os.path.dirname(img_file), exist_ok=True)

        tv_save_image(tensor=tensor, filename=img_file, **image_args)

    def save_image_grid(self, tensor, name, n_iter=None, prefix=False, iter_format="{:05d}", image_args=None):
        """
        Saves images of a 4d- tensor (N, C, H, W) as a image grid into an image file in the image directory of the
        PytorchPlotFileLogger

        Args:
            tensor: 4d- tensor (N, C, H, W)
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method

        """

        if image_args is None:
            image_args = {}

        self.save_image_grid_static(image_dir=self.img_dir, tensor=tensor, name=name, n_iter=n_iter, prefix=prefix,
                                    iter_format=iter_format, image_args=image_args)

    def show_image(self, image, name, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None, **kwargs):
        """
        Calls the save image method (for abstract logger combatibility)

        Args:
            image: Tensor containing the image
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method


        """
        self.save_image(tensor=image, name=name, n_iter=n_iter, iter_format=iter_format, image_args=image_args,
                        prefix=prefix)

    def show_images(self, images, name, n_iter=None, iter_format="{:05d}", prefix=False, image_args=None, **kwargs):
        """
        Calls the save images method (for abstract logger combatibility)

        Args:
            images: List of Tensors
            name: List of file names (corresponding to the images list)
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method

        """
        tensors = {}
        for i, img in enumerate(images):
            tensors[name + "_" + str(i)] = img

        self.save_images(tensors=tensors, n_iter=n_iter, iter_format=iter_format, prefix=prefix, image_args=image_args)

    def show_image_grid(self, images, name, n_iter=None, prefix=False, iter_format="{:05d}", image_args=None,
                        **kwargs):
        """
        Calls the save image grid method (for abstract logger combatibility)

        Args:
            images: 4d- tensor (N, C, H, W)
            name: file-name of the image file
            n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
            iter_format: The format string, which indicates how n_iter will be formated as a string
            prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
            image_args: Arguments for the tensorvision save image method


        """

        self.save_image_grid(tensor=images, name=name, n_iter=n_iter, prefix=prefix, iter_format=iter_format,
                             image_args=image_args)

    @convert_params
    def show_image_grid_heatmap(self, heatmap, background=None, ratio=0.3, normalize=True,
                                colormap=cv2.COLORMAP_JET, name="heatmap", n_iter=None,
                                prefix=False, iter_format="{:05d}", image_args=None, **kwargs):
        """
        Creates heat map from the given map and if given combines it with the background and then
        displays results with as image grid.

        Args:
           heatmap:  4d- tensor (N, C, H, W) to be converted to a heatmap
           background: 4d- tensor (N, C, H, W) background/ context of the heatmap (to be underlayed)
           name: The name of the window
           ratio: The ratio to mix the map with the background (0 = only background, 1 = only map)
           n_iter: The iteration number, formatted with the iter_format and added to the model name (if not None)
           iter_format: The format string, which indicates how n_iter will be formated as a string
           prefix: If True, the formated n_iter will be appended as a prefix, otherwise as a suffix
           image_args: Arguments for the tensorvision save image method

        """

        if image_args is None: image_args = {}

        if n_iter is not None:
            name = name_and_iter_to_filename(name=name, n_iter=n_iter, ending=".png", iter_format=iter_format,
                                             prefix=prefix)
        elif not name.endswith(".png"):
            name += ".png"

        file_name = os.path.join(self.img_dir, name)

        map_grid = np_make_grid(heatmap, normalize=normalize)
        map_ = np.clip(map_grid * 255, a_min=0, a_max=255)
        map_ = map_.astype(np.uint8)

        map_ = cv2.applyColorMap(map_.transpose(1, 2, 0), colormap=colormap)
        map_ = cv2.cvtColor(map_, cv2.COLOR_BGR2RGB)
        map_ = map_.transpose(2, 0, 1)

        fuse_img = map_

        if background is not None:
            img_grid = np_make_grid(background, **image_args)
            image = np.clip(img_grid * 255, a_min=0, a_max=255)
            image = image.astype(np.uint8)

            fuse_img = (1.0 - ratio) * image + ratio * map_

        imsave(file_name, fuse_img.transpose(1, 2, 0))
